"""
Classes for working with Nextmv Cloud Ensemble Runs.

This module provides classes for interacting with ensemble runs in Nextmv Cloud.
It details the core data structures for ensemble definitions.

Classes
-------
RunGroup
    A structure to group execution of child runs for an ensemble run.
RuleObjective
    An enum that specifies the supported evaluation rule objectives.
ToleranceType
    An enum that specifies the supported tolerance types for evaluation rules.
RuleTolerance
    A structure for defining tolerance thresholds for an evaluation rule
EvaluationRule
    A structure to evaluate run results for an ensemble run.
EnsembleDefinition
    Representation of a Nextmv Cloud Ensemble Definition for an application.
"""

from datetime import datetime
from enum import Enum

from nextmv.base_model import BaseModel


class RunGroup(BaseModel):
    """A structure to group child runs for an ensemble run.

    You can import the `RunGroup` class directly from `cloud`:

    ```python
    from nextmv.cloud import RunGroup
    ```

    This class represents a grouping of child runs that share a configuration
    for ensemble run executions.

    Parameters
    ----------
    id : str
        The unique identifier of the run group.
    instance_id : str
        ID of the app instance that this run group executes on.
    options : dict, optional
        Runtime options/parameters for the application.
    repetitions : int, optional
        The number of times the run is to be repeated on the instance and with
        the options defined in the run group
    """

    id: str
    """The unique identifier of the run group."""
    instance_id: str
    """ID of the app instance that this run group executes on."""
    options: dict | None = None
    """Runtime options/parameters for the application."""
    repetitions: int | None = None
    """The number of times the run is to be repeated on the instance and with
    the options defined in the run group"""


class RuleObjective(str, Enum):
    """The value of this data determines how a value of a run is optimized to
    determined which ensemble child run is the "best" for a given metric and
    rule, as well as which other ones are within tolerance of that run for the
    purposes of selecting a result for the ensemble run from among the child runs.

    You can import the `RuleObjective` class directly from `cloud`:

    ```python
    from nextmv.cloud import RuleObjective
    ```

    This enum specifies the supported evaluation rule objectives.

    Attributes
    ----------
    MAXIMIZE : str
        Maximize the value of the evaluated metric.
    MINIMIZE : str
        Minimize the value of the evaluated metric.
    """

    MAXIMIZE = "maximize"
    """Maximize the value of the evaluated metric."""
    MINIMIZE = "minimize"
    """Minimize the value of the evaluated metric."""


class RuleToleranceType(str, Enum):
    """The type of comparison used to determine if a run metric is within
    tolerance of a the "best" run for that rule and metric

    You can import the `RuleToleranceType` class directly from `cloud`:

    ```python
    from nextmv.cloud import RuleToleranceType
    ```

    This enum specifies the supported tolerance types.

    Attributes
    ----------
    ABSOLUTE : str
        Uses the absolute difference between the value of the "best" run and
        the run being evaluated for tolerance
    RELATIVE : str
        Uses the the percentage of the "best" run by which the run being
        evaluted for tolerance differs. A value of `1` is 100%.
    """

    ABSOLUTE = "absolute"
    """Uses the absolute difference between the value of the "best" run and
    the run being evaluated for tolerance"""
    RELATIVE = "relative"
    """Uses the the percentage of the "best" run by which the run being
    evaluted for tolerance differs. A value of `1` is 100%."""


class RuleTolerance(BaseModel):
    """A structure used to determine if a run is within tolerance of of the best
    run (as determined by the objective of the `EvaluationRule` it is defined on).

    You can import the `RuleTolerance` class directly from `cloud`:

    ```python
    from nextmv.cloud import RuleTolerance
    ```

    This class represents the tolerance on a particular evaluation rule by
    which a child run may be selected as the result of an ensemble run.

    value : float
        The value within which runs can deviate from the "best" run
        for that metric to be considered within tolerance of it.
    type : ToleranceType
        The method by which runs are determined to be within tolerance.
    """

    value: float
    """The value within which runs can deviate from the "best" run
    for that metric to be considered within tolerance of it."""
    type: RuleToleranceType
    """The method by which runs are determined to be within tolerance."""


class EvaluationRule(BaseModel):
    """A structure to evaluate run results for an ensemble run.

    You can import the `EvaluationRule` class directly from `cloud`:

    ```python
    from nextmv.cloud import EvaluationRule
    ```

    This class represents a rule by which the child runs for an ensemble run
    will be evaluated for the purpose of selecting an optimal result for the
    ensemble run.

    Parameters
    ----------
    id : str
        The unique identifier of the evaluation rule.
    statistics_path : str
        The path within the statistics of a run output (conforming to Nextmv
        statistics convention and flattened to a string starting with `$` and
        delimited by `.` e.g. `$.result.value`.)
    objective : RuleObjective
        The objective by which runs are optimized for this rule
    tolerance : RuleTolerance
        The tolerance by which runs can be accepted as a potential result
        for an evaluation rule
    index : int, optional
        The index (non-negative integer) of the evalutation rule. Lower indicies
        are evaluated first.
    """

    id: str
    """The unique identifier of the evaluation rule."""
    statistics_path: str
    """The path within the statistics of a run output (conforming to Nextmv
    statistics convention and flattened to a string starting with `$` and
    delimited by `.` e.g. `$.result.value`.)"""
    objective: RuleObjective
    """The objective by which runs are optimized for this rule"""
    tolerance: RuleTolerance
    """The tolerance by which runs can be accepted as a potential result
    for an evaluation rule"""
    index: int
    """The index (non-negative integer) of the evalutation rule. Lower indicies
    are evaluated first."""


class EnsembleDefinition(BaseModel):
    """An ensemble definition for an application.

    You can import the `EnsembleDefinition` class directly from `cloud`:

    ```python
    from nextmv.cloud import EnsembleDefinition
    ```

    A Nextmv Cloud ensemble definition represents a structure by which an
    application can coordinate and execute, and determine the optimal result of
    an ensemble run.

    Parameters
    ----------
    id : str
        The unique identifier of the ensemble definition.
    application_id : str
        ID of the application that this ensemble definition belongs to.
    name : str
        Human-readable name of the ensemble definition.
    description : str
        Detailed description of the ensemble definition.
    run_groups : list[RunGroup], optional
        The run groups that structure the execution of an ensemble run
    rules : list[EvaluationRule], optional
        The rules by which ensemble child runs are evaluated
        to find an optimal result.
    created_at : datetime
        Timestamp when the ensemble definition was created.
    updated_at : datetime
        Timestamp when the ensemble definition was last updated.
    """

    id: str
    """The unique identifier of the ensemble definition."""
    application_id: str
    """ID of the application that this ensemble definition belongs to."""
    name: str = ""
    """Human-readable name of the ensemble definition."""
    description: str = ""
    """Detailed description of the ensemble definition."""
    run_groups: list[RunGroup]
    """The run groups that structure the execution of an ensemble run"""
    rules: list[EvaluationRule]
    """The rules by which ensemble child runs are evaluated
    to find an optimal result."""
    created_at: datetime
    """Timestamp when the ensemble definition was created."""
    updated_at: datetime
    """Timestamp when the ensemble definition was last updated."""



"""Module with the client class.

This module provides the `Client` class for interacting with the Nextmv Cloud
API, and a helper function `get_size` to determine the size of objects.

Classes
-------
Client
    Client that interacts directly with the Nextmv Cloud API.

Functions
---------
get_size(obj)
    Finds the size of an object in bytes.
"""

import os
from dataclasses import dataclass, field
from typing import IO, Any
from urllib.parse import urljoin

import requests
import yaml
from requests.adapters import HTTPAdapter, Retry

from nextmv._serialization import deflated_serialize_json

_MAX_LAMBDA_PAYLOAD_SIZE: int = 500 * 1024 * 1024
"""int: Maximum size of the payload handled by the Nextmv Cloud API.

This constant defines the upper limit for the size of data payloads that can
be sent to the Nextmv Cloud API, specifically for lambda functions. It is set
to 500 MiB.
"""


@dataclass
class Client:
    """
    Client that interacts directly with the Nextmv Cloud API.

    You can import the `Client` class directly from `cloud`:

    ```python
    from nextmv.cloud import Client
    ```

    The API key will be searched, in order of precedence, in:

    1. The `api_key` argument in the constructor.
    2. The `NEXTMV_API_KEY` environment variable.
    3. The `~/.nextmv/config.yaml` file used by the Nextmv CLI.

    Parameters
    ----------
    api_key : str, optional
        API key to use for authenticating with the Nextmv Cloud API. If not
        provided, the client will look for the `NEXTMV_API_KEY` environment
        variable.
    allowed_methods : list[str]
        Allowed HTTP methods to use for retries in requests to the Nextmv
        Cloud API. Defaults to ``["GET", "POST", "PUT", "DELETE"]``.
    backoff_factor : float
        Exponential backoff factor to use for requests to the Nextmv Cloud
        API. Defaults to ``1``.
    backoff_jitter : float
        Jitter to use for requests to the Nextmv Cloud API when backing off.
        Defaults to ``0.1``.
    backoff_max : float
        Maximum backoff time to use for requests to the Nextmv Cloud API, in
        seconds. Defaults to ``60``.
    configuration_file : str
        Path to the configuration file used by the Nextmv CLI. Defaults to
        ``"~/.nextmv/config.yaml"``.
    headers : dict[str, str], optional
        Headers to use for requests to the Nextmv Cloud API. Automatically
        set up with the API key.
    max_retries : int
        Maximum number of retries to use for requests to the Nextmv Cloud
        API. Defaults to ``10``.
    status_forcelist : list[int]
        Status codes to retry for requests to the Nextmv Cloud API. Defaults
        to ``[429]``.
    timeout : float
        Timeout to use for requests to the Nextmv Cloud API, in seconds.
        Defaults to ``20``.
    url : str
        URL of the Nextmv Cloud API. Defaults to
        ``"https://api.cloud.nextmv.io"``.
    console_url : str
        URL of the Nextmv Cloud console. Defaults to
        ``"https://cloud.nextmv.io"``.

    Examples
    --------
    >>> client = Client(api_key="YOUR_API_KEY")
    >>> response = client.request(method="GET", endpoint="/v1/applications")
    >>> print(response.json())
    """

    api_key: str | None = None
    """API key to use for authenticating with the Nextmv Cloud API. If not
    provided, the client will look for the NEXTMV_API_KEY environment
    variable."""
    allowed_methods: list[str] = field(
        default_factory=lambda: ["GET", "POST", "PUT", "DELETE"],
    )
    """Allowed HTTP methods to use for retries in requests to the Nextmv Cloud
    API."""
    backoff_factor: float = 1
    """Exponential backoff factor to use for requests to the Nextmv Cloud
    API."""
    backoff_jitter: float = 0.1
    """Jitter to use for requests to the Nextmv Cloud API when backing off."""
    backoff_max: float = 60
    """Maximum backoff time to use for requests to the Nextmv Cloud API, in
    seconds."""
    configuration_file: str = "~/.nextmv/config.yaml"
    """Path to the configuration file used by the Nextmv CLI."""
    headers: dict[str, str] | None = None
    """Headers to use for requests to the Nextmv Cloud API."""
    max_retries: int = 10
    """Maximum number of retries to use for requests to the Nextmv Cloud
    API."""
    status_forcelist: list[int] = field(
        default_factory=lambda: [429],
    )
    """Status codes to retry for requests to the Nextmv Cloud API."""
    timeout: float = 20
    """Timeout to use for requests to the Nextmv Cloud API."""
    url: str = "https://api.cloud.nextmv.io"
    """URL of the Nextmv Cloud API."""
    console_url: str = "https://cloud.nextmv.io"
    """URL of the Nextmv Cloud console."""

    def __post_init__(self):
        """
        Initializes the client after dataclass construction.

        This method handles the logic for API key retrieval and header
        setup. It checks for the API key in the constructor, environment
        variables, and the configuration file, in that order.

        Raises
        ------
        ValueError
            If `api_key` is an empty string.
            If no API key is found in any of the lookup locations.
            If a profile is specified via `NEXTMV_PROFILE` but not found in
            the configuration file.
            If `apikey` is not found in the configuration file for the
            selected profile.
        """

        if self.api_key is not None and self.api_key != "":
            self._set_headers_api_key(self.api_key)
            return

        if self.api_key == "":
            raise ValueError("api_key cannot be empty")

        api_key_env = os.getenv("NEXTMV_API_KEY")
        if api_key_env is not None:
            self.api_key = api_key_env
            self._set_headers_api_key(api_key_env)
            return

        config_path = os.path.expanduser(self.configuration_file)
        if not os.path.exists(config_path):
            raise ValueError(
                f"no API key set in constructor or NEXTMV_API_KEY env var, and {self.configuration_file} does not exist"
            )

        with open(config_path) as f:
            config = yaml.safe_load(f)

        profile = os.getenv("NEXTMV_PROFILE")
        parent = config
        if profile is not None:
            parent = config.get(profile)
            if parent is None:
                raise ValueError(f"profile {profile} set via NEXTMV_PROFILE but not found in {self.configuration_file}")

        api_key = parent.get("apikey")
        if api_key is None:
            raise ValueError(f"no apiKey found in {self.configuration_file}")
        self.api_key = api_key

        endpoint = parent.get("endpoint")
        if endpoint is not None:
            self.url = f"https://{endpoint}"

        self._set_headers_api_key(api_key)

    def request(
        self,
        method: str,
        endpoint: str,
        data: Any | None = None,
        headers: dict[str, str] | None = None,
        payload: dict[str, Any] | None = None,
        query_params: dict[str, Any] | None = None,
        json_configurations: dict[str, Any] | None = None,
    ) -> requests.Response:
        """
        Makes a request to the Nextmv Cloud API.

        Parameters
        ----------
        method : str
            HTTP method to use (e.g., "GET", "POST").
        endpoint : str
            API endpoint to send the request to (e.g., "/v1/applications").
        data : Any, optional
            Data to send in the request body. Typically used for form data.
            Cannot be used if `payload` is also provided.
        headers : dict[str, str], optional
            Additional headers to send with the request. These will override
            the default client headers if keys conflict.
        payload : dict[str, Any], optional
            JSON payload to send with the request. Prefer using this over
            `data` for JSON requests. Cannot be used if `data` is also
            provided.
        query_params : dict[str, Any], optional
            Query parameters to append to the request URL.
        json_configurations : dict[str, Any], optional
            Additional configurations for JSON serialization. This allows
            customization of the Python `json.dumps` function, such as
            specifying `indent` for pretty printing or `default` for custom
            serialization functions.

        Returns
        -------
        requests.Response
            The response object from the Nextmv Cloud API.

        Raises
        ------
        requests.HTTPError
            If the response status code is not in the 2xx range.
        ValueError
            If both `data` and `payload` are provided.
            If the `payload` size exceeds `_MAX_LAMBDA_PAYLOAD_SIZE`.
            If the `data` size exceeds `_MAX_LAMBDA_PAYLOAD_SIZE`.

        Examples
        --------
        >>> client = Client(api_key="YOUR_API_KEY")
        >>> # Get a list of applications
        >>> response = client.request(method="GET", endpoint="/v1/applications")
        >>> print(response.status_code)
        200
        >>> # Create a new run
        >>> run_payload = {
        ...     "applicationId": "app_id",
        ...     "instanceId": "instance_id",
        ...     "input": {"value": 10}
        ... }
        >>> response = client.request(
        ...     method="POST",
        ...     endpoint="/v1/runs",
        ...     payload=run_payload
        ... )
        >>> print(response.json()["id"])
        run_xxxxxxxxxxxx
        """

        if payload is not None and data is not None:
            raise ValueError("cannot use both data and payload")

        if (
            payload is not None
            and get_size(payload, json_configurations=json_configurations) > _MAX_LAMBDA_PAYLOAD_SIZE
        ):
            raise ValueError(
                f"payload size of {get_size(payload, json_configurations=json_configurations)} bytes exceeds "
                + f"the maximum allowed size of {_MAX_LAMBDA_PAYLOAD_SIZE} bytes"
            )

        if data is not None and get_size(data, json_configurations=json_configurations) > _MAX_LAMBDA_PAYLOAD_SIZE:
            raise ValueError(
                f"data size of {get_size(data, json_configurations=json_configurations)} bytes exceeds "
                + f"the maximum allowed size of {_MAX_LAMBDA_PAYLOAD_SIZE} bytes"
            )

        session = requests.Session()
        retries = Retry(
            total=self.max_retries,
            backoff_factor=self.backoff_factor,
            backoff_jitter=self.backoff_jitter,
            backoff_max=self.backoff_max,
            status_forcelist=self.status_forcelist,
            allowed_methods=self.allowed_methods,
        )
        adapter = HTTPAdapter(max_retries=retries)
        session.mount("https://", adapter)

        kwargs: dict[str, Any] = {
            "url": urljoin(self.url, endpoint),
            "timeout": self.timeout,
        }
        kwargs["headers"] = headers if headers is not None else self.headers
        if data is not None:
            kwargs["data"] = data
        if payload is not None:
            if isinstance(payload, (dict, list)):
                data = deflated_serialize_json(payload, json_configurations=json_configurations)
                kwargs["data"] = data
            else:
                raise ValueError("payload must be a dictionary or a list")
        if query_params is not None:
            kwargs["params"] = query_params

        response = session.request(method=method, **kwargs)

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            raise requests.HTTPError(
                f"request to {endpoint} failed with status code {response.status_code} and message: {response.text}"
            ) from e

        return response

    def upload_to_presigned_url(
        self,
        data: dict[str, Any] | str | None,
        url: str,
        json_configurations: dict[str, Any] | None = None,
        tar_file: str | None = None,
    ) -> None:
        """
        Uploads data to a presigned URL.

        This method is typically used for uploading large input or output files
        directly to cloud storage, bypassing the main API for efficiency.

        Parameters
        ----------
        data : Union[dict[str, Any], str], optional
            The data to upload. If a dictionary is provided, it will be
            JSON-serialized. If a string is provided, it will be uploaded
            as is.
        url : str
            The presigned URL to which the data will be uploaded.
        json_configurations : dict[str, Any], optional
            Additional configurations for JSON serialization. This allows
            customization of the Python `json.dumps` function, such as
            specifying `indent` for pretty printing or `default` for custom
            serialization functions.
        tar_file : str, optional
            If provided, this will be used to upload a tar file instead of
            a JSON string or dictionary. This is useful for uploading large
            files that are already packaged as a tarball. If this is provided,
            `data` is expected to be `None`.

        Raises
        ------
        ValueError
            If `data` is not a dictionary or a string.
        requests.HTTPError
            If the upload request fails.

        Examples
        --------
        Assume `presigned_upload_url` is obtained from a previous API call.
        >>> client = Client(api_key="YOUR_API_KEY")
        >>> input_data = {"value": 42, "items": [1, 2, 3]}
        >>> client.upload_to_presigned_url(data=input_data, url="PRE_SIGNED_URL") # doctest: +SKIP
        """

        upload_data: str | None = None
        if data is not None:
            if isinstance(data, dict):
                upload_data = deflated_serialize_json(data, json_configurations=json_configurations)
            elif isinstance(data, str):
                upload_data = data
            else:
                raise ValueError("data must be a dictionary or a string")

        session = requests.Session()
        retries = Retry(
            total=self.max_retries,
            backoff_factor=self.backoff_factor,
            backoff_jitter=self.backoff_jitter,
            backoff_max=self.backoff_max,
            status_forcelist=self.status_forcelist,
            allowed_methods=self.allowed_methods,
        )
        adapter = HTTPAdapter(max_retries=retries)
        session.mount("https://", adapter)

        kwargs: dict[str, Any] = {
            "url": url,
            "timeout": self.timeout,
        }

        if upload_data is not None:
            kwargs["data"] = upload_data
        elif tar_file is not None and tar_file != "":
            if not os.path.exists(tar_file):
                raise ValueError(f"tar_file {tar_file} does not exist")
            kwargs["data"] = open(tar_file, "rb")
        else:
            raise ValueError("either data or tar_file must be provided")

        response = session.put(**kwargs)

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            raise requests.HTTPError(
                f"upload to presigned URL {url} failed with "
                f"status code {response.status_code} and message: {response.text}"
            ) from e

    def _set_headers_api_key(self, api_key: str) -> None:
        """
        Sets the Authorization and Content-Type headers.

        This is an internal method used to configure the necessary headers
        for API authentication and content type specification.

        Parameters
        ----------
        api_key : str
            The API key to be included in the Authorization header.
        """

        self.headers = {
            "Authorization": f"Bearer {api_key}",
            "Content-Type": "application/json",
        }


def get_size(obj: dict[str, Any] | IO[bytes] | str, json_configurations: dict[str, Any] | None = None) -> int:
    """
    Finds the size of an object in bytes.

    This function supports dictionaries (JSON-serialized), file-like objects
    (by reading their content), and strings.

    Parameters
    ----------
    obj : dict[str, Any] or IO[bytes] or str
        The object whose size is to be determined.
        - If a dict, it's converted to a JSON string.
        - If a file-like object (e.g., opened file), its size is read.
        - If a string, its UTF-8 encoded byte length is calculated.
    json_configurations : dict[str, Any], optional
        Additional configurations for JSON serialization. This allows
        customization of the Python `json.dumps` function, such as specifying
        `indent` for pretty printing or `default` for custom serialization
        functions.

    Returns
    -------
    int
        The size of the object in bytes.

    Raises
    ------
    TypeError
        If the object type is not supported (i.e., not a dict,
        file-like object, or string).

    Examples
    --------
    >>> my_dict = {"key": "value", "number": 123}
    >>> get_size(my_dict)
    30
    >>> import io
    >>> my_string = "Hello, Nextmv!"
    >>> string_io = io.StringIO(my_string)
    >>> # To get size of underlying buffer for StringIO, we need to encode
    >>> string_bytes_io = io.BytesIO(my_string.encode('utf-8'))
    >>> get_size(string_bytes_io)
    14
    >>> get_size("Hello, Nextmv!")
    14
    """

    if isinstance(obj, dict):
        obj_str = deflated_serialize_json(obj, json_configurations=json_configurations)
        return len(obj_str.encode("utf-8"))

    elif hasattr(obj, "read"):
        obj.seek(0, 2)  # Move the cursor to the end of the file
        size = obj.tell()
        obj.seek(0)  # Reset the cursor to the beginning of the file
        return size

    elif isinstance(obj, str):
        return len(obj.encode("utf-8"))

    else:
        raise TypeError("Unsupported type. Only dictionaries, file objects (IO[bytes]), and strings are supported.")



import os
import unittest

from nextmv.model import _cleanup_python_model

import nextmv


class ModelForTesting(nextmv.Model):
    """Dummy decision model for testing purposes."""

    def solve(self, input: nextmv.Input) -> nextmv.Output:
        return nextmv.Output()


class TestModel(unittest.TestCase):
    def tearDown(self):
        """Removes the newly created mlflow elements, while also testing the
        "private" cleanup function."""

        model_configuration = nextmv.ModelConfiguration(
            name="test_model",
        )
        _cleanup_python_model(model_dir=".", model_configuration=model_configuration)

    def test_save(self):
        model = ModelForTesting()
        model_configuration = nextmv.ModelConfiguration(
            name="test_model",
        )
        model.save(model_dir=".", configuration=model_configuration)

        # Assert that the "test_model" directory was created
        model_path = os.path.join(".", model_configuration.name)
        self.assertTrue(os.path.isdir(model_path))

        # Assert that the "test_model" directory is not empty
        self.assertTrue(len(os.listdir(model_path)) > 0)

    @staticmethod
    def _file_name(name: str, relative_location: str = ".") -> str:
        """
        Returns the full path to a file in the current testing directory.

        Parameters
        ----------
        name : str
            The name of the file.
        relative_location : str, optional
            The relative location of the file. The default is ".".

        Returns
        -------
        str
            The full path to the file.
        """

        return os.path.join(os.path.dirname(__file__), relative_location, name)



"""
Runner module for executing local runs.

This module provides functionality to execute local runs.

Functions
---------
run
    Function to execute a local run.
new_run
    Function to initialize a new run.
record_input
    Function to write the input to the appropriate location.
"""

import importlib.util
import json
import os
import shutil
import subprocess
import sys
from datetime import datetime, timezone
from typing import Any

from nextmv.input import INPUTS_KEY
from nextmv.local.local import DEFAULT_INPUT_JSON_FILE, NEXTMV_DIR, RUNS_KEY, calculate_files_size
from nextmv.manifest import Manifest
from nextmv.run import Format, FormatInput, Metadata, RunInformation, StatusV2
from nextmv.safe import safe_id


def run(
    app_id: str,
    src: str,
    manifest: Manifest,
    run_config: dict[str, Any],
    name: str | None = None,
    description: str | None = None,
    input_data: dict[str, Any] | str | None = None,
    inputs_dir_path: str | None = None,
    options: dict[str, Any] | None = None,
) -> str:
    """
    Execute a local run.

    This method recreates, partially, what the Nextmv Cloud does in the backend
    when running an application. A run ID is generated, a run directory is
    created, and the input data is recorded. Then, a subprocess is started to
    execute the application run in a detached manner. This means that the
    application run is not waited upon.

    Parameters
    ----------
    app_id : str
        The ID of the application.
    src : str
        The path to the application source code.
    manifest : Manifest
        The application manifest.
    run_config : dict[str, Any]
        The run configuration.
    name : Optional[str], optional
        The name for the run, by default None.
    description : Optional[str], optional
        The description for the run, by default None.
    input_data : Optional[Union[dict[str, Any], str]], optional
        The input data for the run, by default None. If `inputs_dir_path` is
        provided, this parameter is ignored.
    inputs_dir_path : Optional[str], optional
        The path to the directory containing input files, by default None. If
        provided, this parameter takes precedence over `input_data`.
    options : Optional[dict[str, Any]], optional
        Additional options for the run, by default None.

    Returns
    -------
    str
        The ID of the created run.
    """

    # Check for required optional dependencies
    missing_deps = []
    if importlib.util.find_spec("folium") is None:
        missing_deps.append("folium")
    if importlib.util.find_spec("plotly") is None:
        missing_deps.append("plotly")

    if missing_deps:
        raise ImportError(
            f"{' and '.join(missing_deps)} {'is' if len(missing_deps) == 1 else 'are'} not installed. "
            "Please install optional dependencies with `pip install nextmv[all]`"
        )

    # Initialize the run: create the ID, dir, and write the input.
    run_id = safe_id("local")
    run_dir = new_run(
        app_id=app_id,
        src=src,
        run_id=run_id,
        run_config=run_config,
        name=name,
        description=description,
    )
    record_input(
        run_dir=run_dir,
        run_id=run_id,
        input_data=input_data,
        inputs_dir_path=inputs_dir_path,
    )

    # Start the process as a daemon (detached) so we don't wait for it to
    # finish. We send the input via stdin and close it immediately without
    # waiting. We call the `executor.py` script to do the actual execution.
    stdin_input = json.dumps(
        {
            "run_id": run_id,
            "src": os.path.abspath(src),
            "manifest_dict": manifest.to_dict(),
            "run_dir": os.path.abspath(run_dir),
            "run_config": run_config,
            "input_data": input_data,
            "inputs_dir_path": os.path.abspath(inputs_dir_path) if inputs_dir_path is not None else None,
            "options": options,
        }
    )
    args = [sys.executable, "executor.py"]
    process = subprocess.Popen(
        args,
        env=os.environ,
        text=True,
        stdin=subprocess.PIPE,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        cwd=os.path.dirname(__file__),
        start_new_session=True,  # Detach from parent process
    )
    process.stdin.write(stdin_input)
    process.stdin.close()

    return run_id


def new_run(
    app_id: str,
    src: str,
    run_id: str,
    run_config: dict[str, Any],
    name: str | None = None,
    description: str | None = None,
) -> str:
    """
    Initializes a new run.

    The run information is recorded in a JSON file within the run directory.

    Parameters
    ----------
    app_id : str
        The ID of the application.
    src : str
        The path to the application source code.
    run_id : str
        The ID of the run.
    run_config : dict[str, Any]
        The run configuration.
    name : Optional[str], optional
        The name for the run, by default None.
    description : Optional[str], optional
        The description for the run, by default None.

    Returns
    -------
    str
        The path to the new run directory.
    """

    # First, ensure the runs directory exists.
    runs_dir = os.path.join(src, NEXTMV_DIR, RUNS_KEY)
    os.makedirs(runs_dir, exist_ok=True)

    # Create a new run directory.
    run_dir = os.path.join(runs_dir, run_id)
    os.makedirs(run_dir, exist_ok=True)

    # Create the run information file.
    created_at = datetime.now(timezone.utc)
    metadata = Metadata(
        application_id=app_id,
        application_instance_id="",
        application_version_id="",
        created_at=created_at,
        duration=0.0,
        error="",
        input_size=0.0,
        output_size=0.0,
        format=Format(
            format_input=FormatInput(
                input_type=run_config["format"]["input"]["type"],
            ),
        ),
        status_v2=StatusV2.queued,
    )

    if description is None:
        description = f"Local run created at {created_at.isoformat().replace('+00:00', 'Z')}"

    if name is None:
        name = f"local run {run_id}"

    information = RunInformation(
        description=description,
        id=run_id,
        metadata=metadata,
        name=name,
        user_email="",
    )
    with open(os.path.join(run_dir, f"{run_id}.json"), "w") as f:
        json.dump(information.to_dict(), f, indent=2)

    return run_dir


def record_input(
    run_dir: str,
    run_id: str,
    input_data: dict[str, Any] | str | None = None,
    inputs_dir_path: str | None = None,
) -> None:
    """
    Writes the input to the appropriate location.

    The size of the input is calculated and recorded in the run information.

    Parameters
    ----------
    run_dir : str
        The path to the run directory.
    run_id : str
        The ID of the run.
    input_data : Optional[Union[dict[str, Any], str]], optional
        The input data for the run, by default None. If `inputs_dir_path` is
        provided, this parameter is ignored.
    inputs_dir_path : Optional[str], optional
        The path to the directory containing input files, by default None. If
        provided, this parameter takes precedence over `input_data`.
    """

    # Create the inputs directory.
    run_inputs_dir = os.path.join(run_dir, INPUTS_KEY)
    os.makedirs(run_inputs_dir, exist_ok=True)

    if inputs_dir_path is not None and inputs_dir_path != "":
        # If we specify an inputs directory, we ignore the input_data.
        # Copy all files from inputs_dir_path to run_inputs_dir
        if os.path.exists(inputs_dir_path) and os.path.isdir(inputs_dir_path):
            shutil.copytree(inputs_dir_path, run_inputs_dir, dirs_exist_ok=True)

    elif isinstance(input_data, dict):
        # If no inputs_dir_path is provided, try a single JSON input.
        with open(os.path.join(run_inputs_dir, DEFAULT_INPUT_JSON_FILE), "w") as f:
            json.dump(input_data, f, indent=2)

    elif isinstance(input_data, str):
        # If no inputs_dir_path is provided, try a single TEXT input.
        with open(os.path.join(run_inputs_dir, "input"), "w") as f:
            f.write(input_data)

    else:
        raise ValueError(
            "Invalid input data type: input_data must be a dict or str, or inputs_dir_path must be provided."
        )

    # Update the input size in the run information file.
    calculate_files_size(run_dir, run_id, run_inputs_dir, metadata_key="input_size")



"""
This module defines the configuration list command for the Nextmv CLI.
"""

import typer
from rich.console import Console
from rich.table import Table

from nextmv.cli.configuration.config import API_KEY_KEY, ENDPOINT_KEY, load_config, obscure_api_key
from nextmv.cli.error import error

# Set up subcommand application.
app = typer.Typer()
console = Console()


@app.command()
def list() -> None:
    """
    List the current configuration and all profiles.

    [bold][underline]Examples[/underline][/bold]

    - Show current configuration and all profiles.
        $ [green]nextmv configuration list[/green]
    """

    config = load_config()
    if config == {}:
        error("No configuration found. Please run [code]nextmv configuration[/code].")

    default = {
        "api_key": config.get(API_KEY_KEY),
        "endpoint": config.get(ENDPOINT_KEY),
        "name": "Default",
    }
    profiles = [default]

    for k, v in config.items():
        # Skip default configuration.
        if k in {API_KEY_KEY, ENDPOINT_KEY}:
            continue

        profile = {
            "name": k,
            "api_key": v.get(API_KEY_KEY),
            "endpoint": v.get(ENDPOINT_KEY),
        }
        profiles.append(profile)

    table = Table("Profile name", "API Key", "Endpoint")
    not_set = "[italic]Not set[/italic]"
    for profile in profiles:
        if profile["name"] != "Default":
            table.add_row(
                profile["name"],
                obscure_api_key(profile["api_key"]) if profile.get("api_key") is not None else not_set,
                profile["endpoint"] if profile.get("endpoint") is not None else not_set,
            )
            continue

        api_key = not_set
        if profile.get("api_key") is not None:
            api_key = obscure_api_key(profile["api_key"])

        endpoint = not_set
        if profile.get("endpoint") is not None:
            endpoint = profile["endpoint"]

        table.add_row(
            f"[bold yellow]{profile['name']}[/bold yellow]",
            f"[bold yellow]{api_key}[/bold yellow]",
            f"[bold yellow]{endpoint}[/bold yellow]",
        )
        table.add_section()

    console.print(table)



import csv
import json
import os
import shutil
import unittest
from io import StringIO
from unittest.mock import patch

import pandas as pd

import nextmv


class TestInput(unittest.TestCase):
    """
    Tests for the various classes for loading an input.
    """

    def setUp(self):
        """Set up test fixtures before each test method."""
        # Create test directory and files
        self.test_dir = "test_inputs"
        os.makedirs(self.test_dir, exist_ok=True)

        # Create test JSON file
        self.json_data = {"message": "Hello from JSON", "numbers": [1, 2, 3]}
        with open(f"{self.test_dir}/test_data.json", "w", encoding="utf-8") as f:
            json.dump(self.json_data, f)

        # Create test CSV file
        self.csv_data = [
            {"name": "Alice", "age": "25", "city": "New York"},
            {"name": "Bob", "age": "30", "city": "London"},
            {"name": "Charlie", "age": "35", "city": "Tokyo"},
        ]
        with open(f"{self.test_dir}/test_data.csv", "w", encoding="utf-8", newline="") as f:
            writer = csv.DictWriter(f, fieldnames=["name", "age", "city"])
            writer.writeheader()
            writer.writerows(self.csv_data)

        # Create test text file
        self.text_data = "This is a test text file.\nIt contains multiple lines.\nHello from text!"
        with open(f"{self.test_dir}/test_data.txt", "w", encoding="utf-8") as f:
            f.write(self.text_data)

        # Create test Excel file if pandas is available
        self.excel_data = pd.DataFrame(
            [
                {"product": "Widget A", "price": 10.99, "quantity": 100},
                {"product": "Widget B", "price": 15.50, "quantity": 50},
                {"product": "Widget C", "price": 8.75, "quantity": 200},
            ]
        )
        self.excel_data.to_excel(f"{self.test_dir}/test_data.xlsx", index=False)

    def tearDown(self):
        """Clean up test fixtures after each test method."""
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir)

    def test_local_loader_json_stdin(self):
        sample_input = '{"empanadas": "are_life"}\n'
        input_loader = nextmv.LocalInputLoader()

        with patch("sys.stdin", new=StringIO(sample_input)):
            input_data = input_loader.load()

        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.JSON)
        self.assertEqual(input_data.data, {"empanadas": "are_life"})
        self.assertIsNone(input_data.options)

    def test_local_loader_text_stdin(self):
        sample_input = "empanadas are life\n"
        input_loader = nextmv.LocalInputLoader()

        with patch("sys.stdin", new=StringIO(sample_input)):
            input_data = input_loader.load(input_format=nextmv.InputFormat.TEXT)

        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.TEXT)
        self.assertEqual(input_data.data, "empanadas are life")
        self.assertIsNone(input_data.options)

    def test_local_loader_csv_stdin(self):
        sample_input = '"empanadas","are","life"\n1,2,3\n4,5,6'
        input_loader = nextmv.LocalInputLoader()

        with patch("sys.stdin", new=StringIO(sample_input)):
            input_data = input_loader.load(
                input_format=nextmv.InputFormat.CSV,
                csv_configurations={
                    "quoting": csv.QUOTE_NONNUMERIC,
                },
            )

        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.CSV)
        self.assertEqual(
            list(input_data.data),
            [
                {"empanadas": 1.0, "are": 2.0, "life": 3.0},
                {"empanadas": 4.0, "are": 5.0, "life": 6.0},
            ],
        )
        self.assertIsNone(input_data.options)

    def test_local_loader_with_options(self):
        sample_input = '{"empanadas": "are_life"}\n'
        options = nextmv.Options(nextmv.Option("foo", str, default="bar", required=False))
        input_loader = nextmv.LocalInputLoader()

        with patch("sys.stdin", new=StringIO(sample_input)):
            input_data = input_loader.load(options=options)

        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.JSON)
        self.assertEqual(input_data.data, {"empanadas": "are_life"})
        self.assertIsNotNone(input_data.options)
        self.assertDictEqual(input_data.options.to_dict(), options.to_dict())

    def test_local_loader_json_file(self):
        sample_input = '{"empanadas": "are_life"}'
        input_loader = nextmv.LocalInputLoader()

        with patch("builtins.open", return_value=StringIO(sample_input)):
            input_data = input_loader.load(path="input.json")

        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.JSON)
        self.assertEqual(input_data.data, {"empanadas": "are_life"})
        self.assertIsNone(input_data.options)

    def test_local_loader_text_file(self):
        sample_input = "empanadas are life"
        input_loader = nextmv.LocalInputLoader()

        with patch("builtins.open", return_value=StringIO(sample_input)):
            input_data = input_loader.load(input_format=nextmv.InputFormat.TEXT, path="input.txt")

        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.TEXT)
        self.assertEqual(input_data.data, "empanadas are life")
        self.assertIsNone(input_data.options)

    def test_local_loader_csv_file(self):
        sample_input = '"empanadas","are","life"\n1,2,3\n4,5,6'
        input_loader = nextmv.LocalInputLoader()

        with patch("builtins.open", return_value=StringIO(sample_input)):
            input_data = input_loader.load(
                input_format=nextmv.InputFormat.CSV,
                path="input.csv",
                csv_configurations={
                    "quoting": csv.QUOTE_NONNUMERIC,
                },
            )

        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.CSV)
        self.assertEqual(
            list(input_data.data),
            [
                {"empanadas": 1.0, "are": 2.0, "life": 3.0},
                {"empanadas": 4.0, "are": 5.0, "life": 6.0},
            ],
        )
        self.assertIsNone(input_data.options)

    def test_local_loader_csv_archive_default_dir(self):
        """If the path for loading the input is not provided, the path `input`
        is used for the directory."""
        self._test_local_loader_csv_archive(test_dir="input", load_path="")

        # Should also work if not provided at all.
        self._test_local_loader_csv_archive(test_dir="input", load_path=None)

    def test_local_loader_csv_archive_custom_dir(self):
        """If the path for loading the input is provided, the path is used for
        the directory."""
        self._test_local_loader_csv_archive(test_dir="custom_dir", load_path="custom_dir")

    def _test_local_loader_csv_archive(
        self,
        test_dir: str,
        load_path: str | None = None,
    ):
        """This is an auxiliary function that is used to test the flow of the
        CSV archive input loader but with different directories."""

        # Create the directory if it doesn't exist
        os.makedirs(test_dir, exist_ok=True)

        input_loader = nextmv.LocalInputLoader()

        # Write sample CSV files
        sample_input_1 = '"empanadas","are","life"\n1,2,3\n4,5,6'
        sample_input_2 = '"or","are","tacos"\n7,8,9\n10,11,12'
        with open(f"{test_dir}/empanada_declaration_archive.csv", "w") as file_1:
            file_1.write(sample_input_1)
        with open(f"{test_dir}/taco_declaration_archive.csv", "w") as file_2:
            file_2.write(sample_input_2)

        # Load the CSV archive input
        input_data = input_loader.load(
            nextmv.InputFormat.CSV_ARCHIVE,
            path=load_path,
            csv_configurations={
                "quoting": csv.QUOTE_NONNUMERIC,
            },
        )

        # Do the checks
        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.CSV_ARCHIVE)
        self.assertIsNone(input_data.options)

        self.assertIn(
            "empanada_declaration_archive",
            list(input_data.data.keys()),
        )
        self.assertIn(
            "taco_declaration_archive",
            list(input_data.data.keys()),
        )
        self.assertEqual(
            list(input_data.data["empanada_declaration_archive"]),
            [
                {"empanadas": 1.0, "are": 2.0, "life": 3.0},
                {"empanadas": 4.0, "are": 5.0, "life": 6.0},
            ],
        )
        self.assertEqual(
            list(input_data.data["taco_declaration_archive"]),
            [
                {"or": 7.0, "are": 8.0, "tacos": 9.0},
                {"or": 10.0, "are": 11.0, "tacos": 12.0},
            ],
        )

        # Remove the directory.
        shutil.rmtree(test_dir)

    def test_json_data_file_convenience_function(self):
        """Test the json_data_file convenience function."""
        data_file = nextmv.json_data_file("test_data")

        # Test that the name gets .json extension
        self.assertEqual(data_file.name, "test_data.json")

        # Test that the loader is callable
        self.assertTrue(callable(data_file.loader))

        # Test loading data with the convenience function
        result = data_file.loader(f"{self.test_dir}/test_data.json")
        self.assertEqual(result, self.json_data)

    def test_json_data_file_with_extension(self):
        """Test json_data_file when name already has .json extension."""
        data_file = nextmv.json_data_file("test_data.json")
        self.assertEqual(data_file.name, "test_data.json")

    def test_json_data_file_with_configurations(self):
        """Test json_data_file with JSON-specific configurations."""
        # Create a JSON file with custom format for testing
        custom_json = '{"key": 1.5, "another": 2.7}'
        with open(f"{self.test_dir}/custom.json", "w") as f:
            f.write(custom_json)

        data_file = nextmv.json_data_file("custom", json_configurations={"parse_float": str})
        result = data_file.loader(f"{self.test_dir}/custom.json")

        # With parse_float=str, float values should be strings
        self.assertEqual(result["key"], "1.5")
        self.assertEqual(result["another"], "2.7")

    def test_csv_data_file_convenience_function(self):
        """Test the csv_data_file convenience function."""
        data_file = nextmv.csv_data_file("test_data")

        # Test that the name gets .csv extension
        self.assertEqual(data_file.name, "test_data.csv")

        # Test that the loader is callable
        self.assertTrue(callable(data_file.loader))

        # Test loading data with the convenience function
        result = data_file.loader(f"{self.test_dir}/test_data.csv")
        self.assertEqual(result, self.csv_data)

    def test_csv_data_file_with_extension(self):
        """Test csv_data_file when name already has .csv extension."""
        data_file = nextmv.csv_data_file("test_data.csv")
        self.assertEqual(data_file.name, "test_data.csv")

    def test_csv_data_file_with_configurations(self):
        """Test csv_data_file with CSV-specific configurations."""
        # Create a CSV file with quotes for testing
        quoted_csv = '"name","age","city"\n"Alice","25","New York"\n"Bob","30","London"'
        with open(f"{self.test_dir}/quoted.csv", "w") as f:
            f.write(quoted_csv)

        data_file = nextmv.csv_data_file("quoted", csv_configurations={"quoting": csv.QUOTE_ALL})
        result = data_file.loader(f"{self.test_dir}/quoted.csv")

        expected = [{"name": "Alice", "age": "25", "city": "New York"}, {"name": "Bob", "age": "30", "city": "London"}]
        self.assertEqual(result, expected)

    def test_text_data_file_convenience_function(self):
        """Test the text_data_file convenience function."""
        data_file = nextmv.text_data_file("test_data.txt")

        # Test that the name remains as provided (with extension)
        self.assertEqual(data_file.name, "test_data.txt")

        # Test that the loader is callable
        self.assertTrue(callable(data_file.loader))

        # Test loading data with the convenience function
        result = data_file.loader(f"{self.test_dir}/test_data.txt")
        self.assertEqual(result, self.text_data)

    def test_data_file_class_direct_usage(self):
        """Test creating DataFile instances directly."""

        def custom_loader(file_path):
            with open(file_path, encoding="utf-8") as f:
                return f.read().upper()

        data_file = nextmv.DataFile(name="test_data.txt", loader=custom_loader)

        self.assertEqual(data_file.name, "test_data.txt")
        self.assertTrue(callable(data_file.loader))

        # Test that our custom loader works
        result = data_file.loader(f"{self.test_dir}/test_data.txt")
        self.assertEqual(result, self.text_data.upper())

    def test_data_file_with_excel_pandas(self):
        """Test DataFile with custom Excel loader using pandas."""
        data_file = nextmv.DataFile(
            name="test_data.xlsx",
            loader=lambda file_path, **kwargs: pd.read_excel(file_path, **kwargs),
        )

        self.assertEqual(data_file.name, "test_data.xlsx")

        # Test loading Excel data
        result = data_file.loader(f"{self.test_dir}/test_data.xlsx")

        # Convert to dict for comparison
        result_dict = result.to_dict("records")
        expected_dict = self.excel_data.to_dict("records")
        self.assertEqual(result_dict, expected_dict)

    def test_data_file_with_loader_args_and_kwargs(self):
        """Test DataFile with loader_args and loader_kwargs."""

        def custom_loader(file_path, prefix="", suffix="", multiplier=1):
            with open(file_path, encoding="utf-8") as f:
                content = f.read()
                return f"{prefix}{content * multiplier}{suffix}"

        data_file = nextmv.DataFile(
            name="test_data.txt",
            loader=custom_loader,
            loader_args=["[PREFIX]"],
            loader_kwargs={"suffix": "[SUFFIX]", "multiplier": 2},
        )

        # This test would normally be done through the multi-file loader
        # but here we test the structure is correct
        self.assertEqual(data_file.loader_args, ["[PREFIX]"])
        self.assertEqual(data_file.loader_kwargs, {"suffix": "[SUFFIX]", "multiplier": 2})

    def test_load_multi_file_format(self):
        """Test loading multiple files using MULTI_FILE format."""
        data_files = [
            nextmv.json_data_file("test_data"),
            nextmv.csv_data_file("test_data"),
            nextmv.text_data_file("test_data.txt"),
        ]

        input_data = nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=data_files, path=self.test_dir)

        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.MULTI_FILE)

        # Check that all files were loaded
        self.assertIn("test_data.json", input_data.data)
        self.assertIn("test_data.csv", input_data.data)
        self.assertIn("test_data.txt", input_data.data)

        # Verify data content
        self.assertEqual(input_data.data["test_data.json"], self.json_data)
        self.assertEqual(input_data.data["test_data.csv"], self.csv_data)
        self.assertEqual(input_data.data["test_data.txt"], self.text_data)

    def test_load_multi_file_with_excel(self):
        """Test loading multiple files including Excel using MULTI_FILE format."""
        data_files = [
            nextmv.json_data_file("test_data"),
            nextmv.csv_data_file("test_data"),
            nextmv.text_data_file("test_data.txt"),
            nextmv.DataFile(
                name="test_data.xlsx",
                loader=lambda file_path, **kwargs: pd.read_excel(file_path, **kwargs),
            ),
        ]

        input_data = nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=data_files, path=self.test_dir)

        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.MULTI_FILE)

        # Check that all files were loaded including Excel
        self.assertIn("test_data.json", input_data.data)
        self.assertIn("test_data.csv", input_data.data)
        self.assertIn("test_data.txt", input_data.data)
        self.assertIn("test_data.xlsx", input_data.data)

        # Verify Excel data was loaded as DataFrame
        excel_result = input_data.data["test_data.xlsx"]
        self.assertIsInstance(excel_result, pd.DataFrame)

    def test_load_multi_file_missing_data_files(self):
        """Test that ValueError is raised when data_files is None for MULTI_FILE format."""
        with self.assertRaises(ValueError) as context:
            nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=None)

        self.assertIn("data_files must be provided", str(context.exception))

    def test_load_multi_file_invalid_data_files_type(self):
        """Test that ValueError is raised when data_files is not a list."""
        with self.assertRaises(ValueError) as context:
            nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files="not_a_list")

        self.assertIn("data_files must be a list", str(context.exception))

    def test_load_multi_file_default_directory(self):
        """Test loading multi-file with default directory 'inputs'."""
        # Create inputs directory and copy test files
        inputs_dir = "inputs"
        os.makedirs(inputs_dir, exist_ok=True)

        try:
            # Copy test files to inputs directory
            shutil.copy(f"{self.test_dir}/test_data.json", f"{inputs_dir}/test_data.json")
            shutil.copy(f"{self.test_dir}/test_data.csv", f"{inputs_dir}/test_data.csv")

            data_files = [nextmv.json_data_file("test_data"), nextmv.csv_data_file("test_data")]

            # Load without specifying path (should use default "inputs" directory)
            input_data = nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=data_files)

            self.assertIsInstance(input_data, nextmv.Input)
            self.assertIn("test_data.json", input_data.data)
            self.assertIn("test_data.csv", input_data.data)

        finally:
            if os.path.exists(inputs_dir):
                shutil.rmtree(inputs_dir)

    def test_load_multi_file_nonexistent_directory(self):
        """Test that ValueError is raised when directory doesn't exist."""
        data_files = [nextmv.json_data_file("test_data")]

        with self.assertRaises(ValueError) as context:
            nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=data_files, path="nonexistent_directory")

        self.assertIn("path nonexistent_directory is not a directory", str(context.exception))

    def test_load_multi_file_path_not_directory(self):
        """Test that ValueError is raised when path points to a file instead of directory."""
        # Create a file instead of directory
        test_file = "not_a_directory.txt"
        with open(test_file, "w") as f:
            f.write("test")

        try:
            data_files = [nextmv.json_data_file("test_data")]

            with self.assertRaises(ValueError) as context:
                nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=data_files, path=test_file)

            self.assertIn(f"path {test_file} is not a directory", str(context.exception))

        finally:
            if os.path.exists(test_file):
                os.remove(test_file)

    def test_data_file_loader_with_custom_args_kwargs(self):
        """Test DataFile loader with custom args and kwargs in multi-file context."""

        def custom_json_loader(file_path, prefix="", parse_float=None):
            with open(file_path, encoding="utf-8") as f:
                data = json.load(f, parse_float=parse_float)
                if prefix:
                    return {f"{prefix}_{k}": v for k, v in data.items()}
                return data

        # Create a test JSON file with floats
        test_data = {"value": 3.14, "count": 42}
        with open(f"{self.test_dir}/custom.json", "w") as f:
            json.dump(test_data, f)

        data_file = nextmv.DataFile(
            name="custom.json", loader=custom_json_loader, loader_kwargs={"prefix": "custom", "parse_float": str}
        )

        input_data = nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=[data_file], path=self.test_dir)

        result = input_data.data["custom.json"]
        expected = {"custom_value": "3.14", "custom_count": 42}
        self.assertEqual(result, expected)

    def test_mixed_data_types_multi_file(self):
        """Test loading different data types in a single multi-file input."""
        # Create additional test files
        with open(f"{self.test_dir}/config.txt", "w") as f:
            f.write("debug=true\nverbose=false")

        # Custom configuration parser
        def parse_config(file_path):
            config = {}
            with open(file_path) as f:
                for line in f:
                    key, value = line.strip().split("=")
                    config[key] = value.lower() == "true" if value.lower() in ["true", "false"] else value
            return config

        data_files = [
            nextmv.json_data_file("test_data"),  # Returns dict
            nextmv.csv_data_file("test_data"),  # Returns list of dicts
            nextmv.text_data_file("test_data.txt"),  # Returns string
            nextmv.DataFile(name="config.txt", loader=parse_config),  # Returns dict
        ]

        input_data = nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=data_files, path=self.test_dir)

        # Verify different data types
        self.assertIsInstance(input_data.data["test_data.json"], dict)
        self.assertIsInstance(input_data.data["test_data.csv"], list)
        self.assertIsInstance(input_data.data["test_data.txt"], str)
        self.assertIsInstance(input_data.data["config.txt"], dict)

        # Verify config parsing worked
        self.assertEqual(input_data.data["config.txt"]["debug"], True)
        self.assertEqual(input_data.data["config.txt"]["verbose"], False)

    # Tests for input_data_key functionality
    def test_json_data_file_with_input_data_key(self):
        """Test json_data_file with custom input_data_key."""
        data_file = nextmv.json_data_file("test_data", input_data_key="custom_json_key")

        # Test that input_data_key is set correctly
        self.assertEqual(data_file.input_data_key, "custom_json_key")

        # Test in multi-file context
        input_data = nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=[data_file], path=self.test_dir)

        # Should use custom key instead of filename
        self.assertIn("custom_json_key", input_data.data)
        self.assertNotIn("test_data.json", input_data.data)
        self.assertEqual(input_data.data["custom_json_key"], self.json_data)

    def test_csv_data_file_with_input_data_key(self):
        """Test csv_data_file with custom input_data_key."""
        data_file = nextmv.csv_data_file("test_data", input_data_key="custom_csv_key")

        # Test that input_data_key is set correctly
        self.assertEqual(data_file.input_data_key, "custom_csv_key")

        # Test in multi-file context
        input_data = nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=[data_file], path=self.test_dir)

        # Should use custom key instead of filename
        self.assertIn("custom_csv_key", input_data.data)
        self.assertNotIn("test_data.csv", input_data.data)
        self.assertEqual(input_data.data["custom_csv_key"], self.csv_data)

    def test_text_data_file_with_input_data_key(self):
        """Test text_data_file with custom input_data_key."""
        data_file = nextmv.text_data_file("test_data.txt", input_data_key="custom_text_key")

        # Test that input_data_key is set correctly
        self.assertEqual(data_file.input_data_key, "custom_text_key")

        # Test in multi-file context
        input_data = nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=[data_file], path=self.test_dir)

        # Should use custom key instead of filename
        self.assertIn("custom_text_key", input_data.data)
        self.assertNotIn("test_data.txt", input_data.data)
        self.assertEqual(input_data.data["custom_text_key"], self.text_data)

    def test_data_file_with_input_data_key_direct(self):
        """Test DataFile with custom input_data_key set directly."""

        def custom_loader(file_path):
            with open(file_path, encoding="utf-8") as f:
                return f.read().upper()

        data_file = nextmv.DataFile(name="test_data.txt", loader=custom_loader, input_data_key="custom_direct_key")

        # Test that input_data_key is set correctly
        self.assertEqual(data_file.input_data_key, "custom_direct_key")

        # Test in multi-file context
        input_data = nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=[data_file], path=self.test_dir)

        # Should use custom key instead of filename
        self.assertIn("custom_direct_key", input_data.data)
        self.assertNotIn("test_data.txt", input_data.data)
        self.assertEqual(input_data.data["custom_direct_key"], self.text_data.upper())

    def test_mixed_files_with_and_without_input_data_key(self):
        """Test mix of files with and without custom input_data_key."""
        data_files = [
            nextmv.json_data_file("test_data", input_data_key="json_config"),
            nextmv.csv_data_file("test_data"),  # No custom key, should use filename
            nextmv.text_data_file("test_data.txt", input_data_key="readme_content"),
        ]

        input_data = nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=data_files, path=self.test_dir)

        # Check that custom keys are used where specified
        self.assertIn("json_config", input_data.data)
        self.assertIn("readme_content", input_data.data)
        self.assertNotIn("test_data.json", input_data.data)
        self.assertNotIn("test_data.txt", input_data.data)

        # Check that filename is used when no custom key is specified
        self.assertIn("test_data.csv", input_data.data)

        # Verify data content
        self.assertEqual(input_data.data["json_config"], self.json_data)
        self.assertEqual(input_data.data["test_data.csv"], self.csv_data)
        self.assertEqual(input_data.data["readme_content"], self.text_data)

    def test_input_data_key_with_configurations(self):
        """Test input_data_key works with loader configurations."""
        # Create a CSV file with quotes for testing
        quoted_csv = '"name","age","city"\n"Alice","25","New York"\n"Bob","30","London"'
        with open(f"{self.test_dir}/quoted.csv", "w") as f:
            f.write(quoted_csv)

        data_file = nextmv.csv_data_file(
            "quoted", csv_configurations={"quoting": csv.QUOTE_ALL}, input_data_key="users_data"
        )

        input_data = nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=[data_file], path=self.test_dir)

        # Should use custom key
        self.assertIn("users_data", input_data.data)
        self.assertNotIn("quoted.csv", input_data.data)

        # Verify data was loaded with configurations
        expected = [{"name": "Alice", "age": "25", "city": "New York"}, {"name": "Bob", "age": "30", "city": "London"}]
        self.assertEqual(input_data.data["users_data"], expected)

    def test_input_data_key_with_json_configurations(self):
        """Test input_data_key works with JSON loader configurations."""
        # Create a JSON file with custom format for testing
        custom_json = '{"key": 1.5, "another": 2.7}'
        with open(f"{self.test_dir}/custom.json", "w") as f:
            f.write(custom_json)

        data_file = nextmv.json_data_file(
            "custom", json_configurations={"parse_float": str}, input_data_key="parsed_floats"
        )

        input_data = nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=[data_file], path=self.test_dir)

        # Should use custom key
        self.assertIn("parsed_floats", input_data.data)
        self.assertNotIn("custom.json", input_data.data)

        # With parse_float=str, float values should be strings
        self.assertEqual(input_data.data["parsed_floats"]["key"], "1.5")
        self.assertEqual(input_data.data["parsed_floats"]["another"], "2.7")

    def test_input_data_key_none_uses_filename(self):
        """Test that when input_data_key is None, filename is used as key."""
        data_file = nextmv.json_data_file("test_data", input_data_key=None)

        # Test that input_data_key is None
        self.assertIsNone(data_file.input_data_key)

        # Test in multi-file context
        input_data = nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=[data_file], path=self.test_dir)

        # Should use filename as key
        self.assertIn("test_data.json", input_data.data)
        self.assertEqual(input_data.data["test_data.json"], self.json_data)

    def test_duplicate_input_data_keys_raises_error(self):
        """Test that ValueError is raised when multiple files have the same input_data_key."""
        # Create second JSON file
        second_json_data = {"different": "content", "values": [4, 5, 6]}
        with open(f"{self.test_dir}/second_data.json", "w", encoding="utf-8") as f:
            json.dump(second_json_data, f)

        data_files = [
            nextmv.json_data_file("test_data", input_data_key="shared_key"),
            nextmv.json_data_file("second_data", input_data_key="shared_key"),
        ]

        # Should raise ValueError for duplicate keys
        with self.assertRaises(ValueError) as context:
            nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=data_files, path=self.test_dir)

        self.assertIn("Duplicate input data key found: shared_key", str(context.exception))

    def test_duplicate_key_filename_conflict_raises_error(self):
        """Test that ValueError is raised when custom input_data_key conflicts with a filename."""
        data_files = [
            nextmv.json_data_file("test_data"),  # Uses filename "test_data.json" as key
            nextmv.csv_data_file("test_data", input_data_key="test_data.json"),  # Custom key conflicts with filename
        ]

        # Should raise ValueError for duplicate keys
        with self.assertRaises(ValueError) as context:
            nextmv.load(nextmv.InputFormat.MULTI_FILE, data_files=data_files, path=self.test_dir)

        self.assertIn("Duplicate input data key found: test_data.json", str(context.exception))



"""
This module defines the community clone command for the Nextmv CLI.
"""

import os
import shutil
import tarfile
import tempfile
from collections.abc import Callable
from typing import Annotated

import rich
import typer

from nextmv.cli.community.list import download_file, download_manifest, find_app, versions_table
from nextmv.cli.error import error
from nextmv.cli.options import ProfileOption

# Set up subcommand application.
app = typer.Typer()

# Helpful constants.
LATEST_VERSION = "latest"


@app.command()
def clone(
    app: Annotated[
        str,
        typer.Option("--app", "-a", help="The name of the community app to clone.", metavar="COMMUNITY_APP"),
    ],
    directory: Annotated[
        str | None,
        typer.Option(
            "--directory",
            "-d",
            help="The directory in which to clone the app. Default is the name of the app at current directory.",
            metavar="DIRECTORY",
        ),
    ] = None,
    version: Annotated[
        str | None,
        typer.Option(
            "--version",
            "-v",
            help="The version of the community app to clone.",
            metavar="VERSION",
        ),
    ] = LATEST_VERSION,
    profile: ProfileOption = None,
) -> None:
    """
    Clone a community app locally.

    By default, the [magenta]latest[/magenta] version will be used. You can
    specify a version with the [code]--version[/code] flag, and customize the
    output directory with the [code]--directory[/code] flag. If you want to
    list the available apps, use the [code]nextmv community list[/code] command.

    [bold][underline]Examples[/underline][/bold]

    - Clone the [magenta]go-nextroute[/magenta] community app (under the
      [magenta]"go-nextroute"[/magenta] directory), using the [magenta]latest[/magenta] version.
        $ [green]nextmv community clone --app go-nextroute[/green]

    - Clone the [magenta]go-nextroute[/magenta] community app under the
      [magenta]"~/sample/my_app"[/magenta] directory, using the [magenta]latest[/magenta] version.
        $ [green]nextmv community clone --app go-nextroute --directory ~/sample/my_app[/green]

    - Clone the [magenta]go-nextroute[/magenta] community app (under the
      [magenta]"go-nextroute"[/magenta] directory), using version [magenta]v1.2.0[/magenta].
        $ [green]nextmv community clone --app go-nextroute --version v1.2.0[/green]

    - Clone the [magenta]go-nextroute[/magenta] community app (under the
      [magenta]"go-nextroute"[/magenta] directory), using the [magenta]latest[/magenta] version
      and a profile named [magenta]hare[/magenta].
        $ [green]nextmv community clone --app go-nextroute --profile hare[/green]
    """

    manifest = download_manifest(profile=profile)
    app_obj = find_app(manifest, app)

    if version is not None and version == "":
        error("The [code]--version[/code] flag cannot be an empty string.")

    if not app_has_version(app_obj, version):
        # We don't use error() here to allow printing something before exiting.
        rich.print(
            f"[red]Error:[/red] Version [magenta]{version}[/magenta] not found "
            f"for community app [magenta]{app}[/magenta]. Available versions are:"
        )
        versions_table(manifest, app)

        raise typer.Exit(code=1)

    original_version = version
    if version == LATEST_VERSION:
        version = app_obj.get("latest_app_version")

    destination = directory
    if directory is None or directory == "":
        destination = app

    full_destination = get_valid_path(destination, os.stat)
    os.makedirs(full_destination, exist_ok=True)

    tarball = f"{app}_{version}.tar.gz"
    s3_file_path = f"{app}/{version}/{tarball}"
    downloaded_object = download_object(
        file=s3_file_path,
        path="community-apps",
        output_dir=full_destination,
        output_file=tarball,
        profile=profile,
    )

    # Extract the tarball to a temporary directory to handle nested structure
    with tempfile.TemporaryDirectory() as temp_dir:
        with tarfile.open(downloaded_object, "r:gz") as tar:
            tar.extractall(path=temp_dir)

        # Find the extracted directory (typically the app name)
        extracted_items = os.listdir(temp_dir)
        if len(extracted_items) == 1 and os.path.isdir(os.path.join(temp_dir, extracted_items[0])):
            # Move contents from the extracted directory to full_destination
            extracted_dir = os.path.join(temp_dir, extracted_items[0])
            for item in os.listdir(extracted_dir):
                shutil.move(os.path.join(extracted_dir, item), full_destination)
        else:
            # If structure is unexpected, move everything directly
            for item in extracted_items:
                shutil.move(os.path.join(temp_dir, item), full_destination)

    # Remove the tarball after extraction
    os.remove(downloaded_object)

    rich.print(
        f":white_check_mark: Successfully cloned the [magenta]{app}[/magenta] community app, "
        f"using version [magenta]{original_version}[/magenta] in path: [magenta]{full_destination}[/magenta]."
    )


def app_has_version(app_obj: dict, version: str) -> bool:
    """
    Check if the given app object has the specified version.

    Parameters
    ----------
    app_obj : dict
        The community app object.
    version : str
        The version to check.

    Returns
    -------
    bool
        True if the app has the specified version, False otherwise.
    """

    if version == LATEST_VERSION:
        version = app_obj.get("latest_app_version")

    if version in app_obj.get("app_versions", []):
        return True

    return False


def get_valid_path(path: str, stat_fn: Callable[[str], os.stat_result], ending: str = "") -> str:
    """
    Validates and returns a non-existing path. If the path exists,
    it will append a number to the path and return it. If the path does not
    exist, it will return the path as is.

    The ending parameter is used to check if the path ends with a specific
    string. This is useful to specify if it is a file (like foo.json, in which
    case the next iteration is foo-1.json) or a directory (like foo, in which
    case the next iteration is foo-1).

    Parameters
    ----------
    path : str
        The initial path to validate.
    stat_fn : Callable[[str], os.stat_result]
        A function that takes a path and returns its stat result.
    ending : str, optional
        The expected ending of the path (e.g., file extension), by default "".

    Returns
    -------
    str
        A valid, non-existing path.

    Raises
    ------
    Exception
        If an unexpected error occurs during path validation
    """
    base_name = os.path.basename(path)
    name_without_ending = base_name.removesuffix(ending) if ending else base_name

    while True:
        try:
            stat_fn(path)
            # If we get here, the path exists
            # Get folder/file name number, increase it and create new path
            name = os.path.basename(path)

            # Get folder/file name number
            parts = name.split("-")
            last = parts[-1].removesuffix(ending) if ending else parts[-1]

            # Save last folder name index to be changed
            i = path.rfind(name)

            try:
                num = int(last)
                # Increase number and create new path
                if ending:
                    temp_path = path[:i] + f"{name_without_ending}-{num + 1}{ending}"
                else:
                    temp_path = path[:i] + f"{base_name}-{num + 1}"
                path = temp_path
            except ValueError:
                # If there is no number, add it
                if ending:
                    temp_path = path[:i] + f"{name_without_ending}-1{ending}"
                else:
                    temp_path = path[:i] + f"{name}-1"
                path = temp_path

        except FileNotFoundError:
            # Path doesn't exist, we can use it
            return path
        except Exception:
            # Re-raise unexpected errors
            error(f"An unexpected error occurred while validating the path: {path}")


def download_object(file: str, path: str, output_dir: str, output_file: str, profile: str | None = None) -> str:
    """
    Downloads an object from the internal bucket and saves it to the specified
    output directory.

    Parameters
    ----------
    file : str
        The name of the file to download.
    path : str
        The directory in the bucket where the file is located.
    output_dir : str
        The local directory where the file will be saved.
    output_file : str
        The name of the output file.
    profile : str | None
        The profile name to use. If None, the default profile is used.

    Returns
    -------
    str
        The path to the downloaded file.
    """

    response = download_file(directory=path, file=file, profile=profile)
    file_name = os.path.join(output_dir, output_file)

    with open(file_name, "wb") as f:
        f.write(response.content)

    return file_name



import os
import unittest

from nextmv.cloud import Client


class TestClient(unittest.TestCase):
    def test_api_key(self):
        client1 = Client(api_key="foo")
        self.assertEqual(client1.api_key, "foo")

        os.environ["NEXTMV_API_KEY"] = "bar"
        client2 = Client()
        self.assertEqual(client2.api_key, "bar")
        os.environ.pop("NEXTMV_API_KEY")

        with self.assertRaises(ValueError):
            Client(api_key="")

        with self.assertRaises(ValueError):
            Client(configuration_file="")

        os.environ["NEXTMV_PROFILE"] = "i-like-turtles"
        with self.assertRaises(ValueError):
            Client()

    def test_headers(self):
        client1 = Client(api_key="foo")
        self.assertIsNotNone(client1.headers)

        os.environ["NEXTMV_API_KEY"] = "bar"
        client2 = Client()
        self.assertEqual(client2.api_key, "bar")
        self.assertIsNotNone(client2.headers)
        os.environ.pop("NEXTMV_API_KEY")



import unittest

from nextmv.manifest import (
    Manifest,
    ManifestContent,
    ManifestContentMultiFile,
    ManifestContentMultiFileInput,
    ManifestContentMultiFileOutput,
    ManifestOption,
    ManifestOptions,
    ManifestOptionUI,
    ManifestPython,
    ManifestPythonArch,
    ManifestRuntime,
    ManifestType,
    ManifestValidation,
)
from nextmv.model import ModelConfiguration
from nextmv.options import Option, Options, OptionsEnforcement


class TestManifest(unittest.TestCase):
    def test_from_model_configuration(self):
        options = Options(
            Option("param1", str, ""),
            Option("param2", str, ""),
        )
        model_configuration = ModelConfiguration(
            name="super_cool_model",
            requirements=[
                "one_requirement",
                "another_requirement",
            ],
            options=options,
        )
        manifest = Manifest.from_model_configuration(model_configuration)

        self.assertListEqual(
            manifest.files,
            ["main.py", f"{model_configuration.name}/**"],
        )
        self.assertEqual(manifest.runtime, ManifestRuntime.PYTHON)
        self.assertEqual(manifest.type, ManifestType.PYTHON)

        manifest_python = ManifestPython.from_dict(
            {
                "pip-requirements": "model_requirements.txt",
                "model": {
                    "name": model_configuration.name,
                    "options": model_configuration.options.options_dict(),
                },
            }
        )
        self.assertEqual(manifest.python, manifest_python)
        self.assertEqual(manifest_python.pip_requirements, "model_requirements.txt")

    def test_from_model_configuration_with_validation(self):
        options = Options(
            Option("param1", str, "default_value", "A description", True),
            Option("param2", bool, True, "A description", True),
        )

        validation_config = OptionsEnforcement(
            strict=True,
            validation_enforce=True,
        )

        model_configuration = ModelConfiguration(
            name="super_cool_model",
            requirements=[
                "one_requirement",
                "another_requirement",
            ],
            options=options,
            options_enforcement=validation_config,
        )
        manifest = Manifest.from_model_configuration(model_configuration)

        self.assertListEqual(
            manifest.files,
            ["main.py", f"{model_configuration.name}/**"],
        )
        self.assertEqual(manifest.runtime, ManifestRuntime.PYTHON)
        self.assertEqual(manifest.type, ManifestType.PYTHON)

        manifest_python = ManifestPython.from_dict(
            {
                "pip-requirements": "model_requirements.txt",
                "model": {
                    "name": model_configuration.name,
                    "options": model_configuration.options.options_dict(),
                },
            }
        )
        self.assertEqual(manifest.python, manifest_python)
        self.assertEqual(manifest_python.pip_requirements, "model_requirements.txt")
        self.assertEqual(manifest.configuration.options.strict, validation_config.strict)
        self.assertEqual(manifest.configuration.options.validation.enforce, "all")
        self.assertEqual(manifest.configuration.options.items, ManifestOptions.from_options(options).items)

    def test_manifest_python_from_dict(self):
        manifest_python_dict = {
            "pip-requirements": "foo_requirements.txt",
            "version": 3.11,
            "arch": "amd64",
            "model": {
                "name": "foo_model",
            },
        }

        manifest_python = ManifestPython.from_dict(manifest_python_dict)

        self.assertEqual(manifest_python.pip_requirements, "foo_requirements.txt")
        self.assertEqual(manifest_python.version, "3.11")
        self.assertEqual(manifest_python.arch, ManifestPythonArch.AMD64)
        self.assertEqual(manifest_python.model.name, "foo_model")

    def test_manifest_python_direct_instantiation(self):
        manifest_python = ManifestPython(
            pip_requirements="foo_requirements.txt",
            model={"name": "foo_model"},
        )

        self.assertEqual(manifest_python.pip_requirements, "foo_requirements.txt")
        self.assertEqual(manifest_python.model.name, "foo_model")

    def test_manifest_from_yaml(self):
        manifest = Manifest.from_yaml("tests/cloud")

        self.assertListEqual(
            manifest.files,
            ["main.py", "super_cool_model/**"],
        )

        self.assertEqual(manifest.runtime, ManifestRuntime.PYTHON)

        self.assertEqual(manifest.type, ManifestType.PYTHON)

        self.assertEqual(manifest.python.pip_requirements, "model_requirements.txt")
        self.assertEqual(manifest.python.model.name, "super_cool_model")
        self.assertListEqual(
            manifest.python.model.options,
            [
                {
                    "default": 30,
                    "description": "Max runtime duration (in seconds).",
                    "name": "duration",
                    "param_type": "<class 'int'>",
                    "required": False,
                },
            ],
        )

        self.assertEqual(manifest.pre_push, "echo 'hello world - pre-push'")

        self.assertEqual(manifest.build.command, "echo 'hello world - build'")
        self.assertDictEqual(
            manifest.build.environment,
            {
                "SUPER": "COOL",
                "EXTRA": "AWESOME",
            },
        )
        self.assertEqual(manifest.configuration.options.format, ["-{{name}}", "{{value}}"])

        self.assertDictEqual(
            manifest.configuration.content.to_dict(),
            {
                "format": "multi-file",
                "multi-file": {
                    "input": {
                        "path": "my-inputs",
                    },
                    "output": {
                        "statistics": "my-outputs/statistics.json",
                        "assets": "my-outputs/assets.json",
                        "solutions": "my-outputs/solutions",
                    },
                },
            },
        )

    def test_extract_options(self):
        manifest = Manifest.from_yaml("tests/cloud")
        options = manifest.extract_options(should_parse=False)
        self.assertEqual(len(options.options), 5)

        found = {
            "string": False,
            "bool": False,
            "int": False,
            "float": False,
        }

        for option in options.options:
            if option.option_type is str:
                found["string"] = True
            elif option.option_type is bool:
                found["bool"] = True
            elif option.option_type is int:
                found["int"] = True
            elif option.option_type is float:
                found["float"] = True

        self.assertTrue(found["string"])
        self.assertTrue(found["bool"])
        self.assertTrue(found["int"])
        self.assertTrue(found["float"])

        self.assertEqual(options.options[3].display_name, "a float parameter")

        self.assertEqual(options.options[4].control_type, "select")
        self.assertEqual(options.options[4].hidden_from, ["operator"])

        manifest2 = Manifest(
            files=["main.py"],
        )
        options2 = manifest2.extract_options()
        self.assertIsNone(options2)

    def test_from_options(self):
        options = Options(
            Option("param1", str, "default", "A description", True),
            Option("param2", bool, True, "A description", True),
            Option("param3", int, 42, "A description", True),
            Option("param4", float, 3.14, "A description", True),
        )
        manifest = Manifest.from_options(options)

        self.assertListEqual(manifest.files, ["main.py"])
        self.assertEqual(manifest.runtime, ManifestRuntime.PYTHON)
        self.assertEqual(manifest.type, ManifestType.PYTHON)
        self.assertEqual(manifest.python.pip_requirements, "requirements.txt")
        self.assertEqual(manifest.configuration.options.strict, False)
        self.assertEqual(manifest.configuration.options.validation, ManifestValidation(enforce="none"))
        self.assertListEqual(
            manifest.configuration.options.items,
            [
                ManifestOption(
                    name="param1",
                    option_type="string",
                    default="default",
                    description="A description",
                    required=True,
                    ui=None,
                ),
                ManifestOption(
                    name="param2", option_type="bool", default=True, description="A description", required=True, ui=None
                ),
                ManifestOption(
                    name="param3", option_type="int", default=42, description="A description", required=True, ui=None
                ),
                ManifestOption(
                    name="param4",
                    option_type="float",
                    default=3.14,
                    description="A description",
                    required=True,
                    ui=None,
                ),
            ],
        )

    def test_manifest_options_from_options(self):
        options = Options(
            Option("param1", str, "default", "A description", True),
        )
        manifest_options = ManifestOptions.from_options(options, format=["-{{name}}", "{{value}}"])
        self.assertEqual(manifest_options.format, ["-{{name}}", "{{value}}"])
        self.assertEqual(manifest_options.strict, False)

    def test_manifest_from_dict(self):
        manifest_dict = {
            "type": "go",
            "runtime": "ghcr.io/nextmv-io/runtime/default:latest",
            "files": ["./build/binary"],
            "execution": {
                "entrypoint": "./binary",
                "cwd": "./build/",
            },
        }

        manifest = Manifest.from_dict(manifest_dict)

        self.assertEqual(manifest.type, ManifestType.GO)
        self.assertEqual(manifest.runtime, ManifestRuntime.DEFAULT)
        self.assertListEqual(manifest.files, ["./build/binary"])
        self.assertEqual(manifest.execution.entrypoint, "./binary")
        self.assertEqual(manifest.execution.cwd, "./build/")

    def test_manifest_content_from_dict(self):
        manifest_content_dict = {
            "format": "multi-file",
            "multi-file": {
                "input": {
                    "path": "data/input_data",
                },
                "output": {
                    "statistics": "data/output/stats.json",
                    "assets": "data/output/assets.json",
                    "solutions": "data/output/solutions",
                },
            },
        }

        manifest_content = ManifestContent.from_dict(manifest_content_dict)

        self.assertEqual(manifest_content.format, "multi-file")
        self.assertIsInstance(manifest_content.multi_file, ManifestContentMultiFile)
        self.assertIsInstance(manifest_content.multi_file.input, ManifestContentMultiFileInput)
        self.assertIsInstance(manifest_content.multi_file.output, ManifestContentMultiFileOutput)
        self.assertEqual(manifest_content.multi_file.input.path, "data/input_data")
        self.assertEqual(manifest_content.multi_file.output.statistics, "data/output/stats.json")
        self.assertEqual(manifest_content.multi_file.output.assets, "data/output/assets.json")
        self.assertEqual(manifest_content.multi_file.output.solutions, "data/output/solutions")

    def test_from_options_with_validation(self):
        options = Options(
            Option(
                "param1",
                str,
                "default",
                "A description",
                True,
                additional_attributes={"max_length": 100},
                control_type="input",
            ),
            Option("param2", bool, True, "A description", True),
            Option("param3", int, 42, "A description", True, additional_attributes={"min": 0, "max": 100, "step": 1}),
            Option("param4", float, 3.14, "A description", True, display_name="a float parameter"),
            Option(
                "param5",
                str,
                "default",
                "A description",
                True,
                additional_attributes={"values": ["option1", "option2"]},
                control_type="select",
                hidden_from=["operator"],
            ),
        )
        manifest = Manifest.from_options(options, OptionsEnforcement(strict=True, validation_enforce=True))

        self.assertListEqual(manifest.files, ["main.py"])
        self.assertEqual(manifest.runtime, ManifestRuntime.PYTHON)
        self.assertEqual(manifest.type, ManifestType.PYTHON)
        self.assertEqual(manifest.python.pip_requirements, "requirements.txt")
        self.assertEqual(manifest.configuration.options.strict, True)
        self.assertEqual(manifest.configuration.options.validation, ManifestValidation(enforce="all"))
        self.assertListEqual(
            manifest.configuration.options.items,
            [
                ManifestOption(
                    name="param1",
                    option_type="string",
                    default="default",
                    description="A description",
                    required=True,
                    additional_attributes={"max_length": 100},
                    ui=ManifestOptionUI(control_type="input"),
                ),
                ManifestOption(
                    name="param2",
                    option_type="bool",
                    default=True,
                    description="A description",
                    required=True,
                ),
                ManifestOption(
                    name="param3",
                    option_type="int",
                    default=42,
                    description="A description",
                    required=True,
                    additional_attributes={"min": 0, "max": 100, "step": 1},
                    ui=None,
                ),
                ManifestOption(
                    name="param4",
                    option_type="float",
                    default=3.14,
                    description="A description",
                    required=True,
                    ui=ManifestOptionUI(display_name="a float parameter"),
                ),
                ManifestOption(
                    name="param5",
                    option_type="string",
                    default="default",
                    description="A description",
                    required=True,
                    additional_attributes={"values": ["option1", "option2"]},
                    ui=ManifestOptionUI(control_type="select", hidden_from=["operator"]),
                ),
            ],
        )


class TestManifestOption(unittest.TestCase):
    def test_from_option(self):
        test_cases = [
            {
                "name": "string option",
                "option": Option("param1", str, "default", "A description", True),
                "expected_option_type": "string",
            },
            {
                "name": "bool option",
                "option": Option("param2", bool, True, "A description", True),
                "expected_option_type": "bool",
            },
            {
                "name": "int option",
                "option": Option("param3", int, 42, "A description", True),
                "expected_option_type": "int",
            },
            {
                "name": "float option",
                "option": Option("param4", float, 3.14, "A description", True),
                "expected_option_type": "float",
            },
        ]

        for test_case in test_cases:
            with self.subTest(test_case["name"]):
                option = test_case["option"]
                manifest_option = ManifestOption.from_option(option)

                self.assertEqual(manifest_option.name, option.name)
                self.assertEqual(manifest_option.option_type, test_case["expected_option_type"])
                self.assertEqual(manifest_option.default, option.default)
                self.assertEqual(manifest_option.description, option.description)
                self.assertEqual(manifest_option.required, option.required)

    def test_to_option(self):
        test_cases = [
            {
                "name": "string option",
                "manifest_option": ManifestOption(
                    name="param1",
                    option_type="string",
                    default="default",
                    description="A description",
                    required=True,
                ),
                "expected_option_type": str,
            },
            {
                "name": "bool option",
                "manifest_option": ManifestOption(
                    name="param2",
                    option_type="bool",
                    default=True,
                    description="A description",
                    required=True,
                ),
                "expected_option_type": bool,
            },
            {
                "name": "int option",
                "manifest_option": ManifestOption(
                    name="param3",
                    option_type="int",
                    default=42,
                    description="A description",
                    required=True,
                ),
                "expected_option_type": int,
            },
            {
                "name": "float option",
                "manifest_option": ManifestOption(
                    name="param4",
                    option_type="float",
                    default=3.14,
                    description="A description",
                    required=True,
                ),
                "expected_option_type": float,
            },
        ]

        for test_case in test_cases:
            with self.subTest(test_case["name"]):
                manifest_option = test_case["manifest_option"]
                option = manifest_option.to_option()

                self.assertEqual(option.name, manifest_option.name)
                self.assertIs(option.option_type, test_case["expected_option_type"])
                self.assertEqual(option.default, manifest_option.default)
                self.assertEqual(option.description, manifest_option.description)
                self.assertEqual(option.required, manifest_option.required)



"""
This module defines the community command tree for the Nextmv CLI.
"""

import typer

from nextmv.cli.community.clone import app as clone_app
from nextmv.cli.community.list import app as list_app

# Set up subcommand application.
app = typer.Typer()
app.add_typer(list_app)
app.add_typer(clone_app)


@app.callback()
def callback() -> None:
    """
    Interact with community apps, which are pre-built decision models.

    Community apps are maintained in the following GitHub repository:
    [link=https://github.com/nextmv-io/community-apps][bold]nextmv-io/community-apps[/bold][/link].
    """
    pass



"""
Unit tests for the nextmv.local.runner module.
"""

import json
import os
import shutil
import sys
import tempfile
import unittest
from unittest.mock import Mock, patch

from nextmv.local.local import NEXTMV_DIR, RUNS_KEY
from nextmv.local.runner import new_run, record_input, run
from nextmv.manifest import Manifest, ManifestRuntime


class TestLocalRunner(unittest.TestCase):
    """Test cases for the local runner module."""

    def setUp(self):
        """Set up test fixtures."""
        self.test_dir = tempfile.mkdtemp()
        self.test_src = os.path.join(self.test_dir, "test_app")
        os.makedirs(self.test_src)

        # Create a basic manifest file
        self.manifest_content = {
            "spec_version": "v1beta1",
            "id": "test-app",
            "name": "Test App",
            "description": "Test application",
            "execution": {"entrypoint": "main.py"},
            "type": "python",
            "runtime": "ghcr.io/nextmv-io/runtime/python:3.11",
            "files": ["main.py"],
        }

        # Create a basic entrypoint file
        self.entrypoint_content = """
import json
import sys

# Read input from stdin
try:
    input_data = json.load(sys.stdin)
except:
    input_data = {"test": "data"}

# Simple output
output = {
    "solution": {"result": 42},
    "statistics": {"duration": 0.1},
    "assets": []
}

print(json.dumps(output))
"""

        with open(os.path.join(self.test_src, "main.py"), "w") as f:
            f.write(self.entrypoint_content)

    def tearDown(self):
        """Clean up test fixtures."""
        shutil.rmtree(self.test_dir, ignore_errors=True)

    def test_new_run_creates_directory_structure(self):
        """Test that new_run creates the proper directory structure."""
        run_id = "test-run-123"
        run_config = {"format": {"input": {"type": "json"}, "output": {"type": "json"}}}

        result_dir = new_run(app_id="sample-app", src=self.test_src, run_id=run_id, run_config=run_config)

        expected_dir = os.path.join(self.test_src, NEXTMV_DIR, RUNS_KEY, run_id)
        self.assertEqual(result_dir, expected_dir)
        self.assertTrue(os.path.exists(expected_dir))
        self.assertTrue(os.path.isdir(expected_dir))

    def test_new_run_creates_runs_dir_if_not_exists(self):
        """Test that new_run creates the runs directory if it doesn't exist."""
        run_id = "test-run-456"
        run_config = {"format": {"input": {"type": "json"}, "output": {"type": "json"}}}

        # Ensure .nextmv/runs doesn't exist
        runs_dir = os.path.join(self.test_src, NEXTMV_DIR, RUNS_KEY)
        self.assertFalse(os.path.exists(runs_dir))

        result_dir = new_run(app_id="sample-app", src=self.test_src, run_id=run_id, run_config=run_config)

        self.assertTrue(os.path.exists(runs_dir))
        self.assertTrue(os.path.exists(result_dir))

    def test_record_input_with_dict_data(self):
        """Test record_input with dictionary input data."""
        run_dir = os.path.join(self.test_dir, "test_run")
        os.makedirs(run_dir)

        input_data = {"test": "data", "value": 42}

        # Create minimal metadata file that calculate_files_size expects
        metadata_file = os.path.join(run_dir, "test_run.json")
        with open(metadata_file, "w") as f:
            json.dump({"metadata": {}}, f)

        record_input(run_dir, run_id="test_run", input_data=input_data)

        # Check that inputs directory was created
        inputs_dir = os.path.join(run_dir, "inputs")
        self.assertTrue(os.path.exists(inputs_dir))

        # Check that input.json was created with correct content
        input_file = os.path.join(inputs_dir, "input.json")
        self.assertTrue(os.path.exists(input_file))

        with open(input_file) as f:
            saved_data = json.load(f)

        self.assertEqual(saved_data, input_data)

    def test_record_input_with_string_data(self):
        """Test record_input with string input data."""
        run_dir = os.path.join(self.test_dir, "test_run")
        os.makedirs(run_dir)

        input_data = "test string input"

        # Create minimal metadata file that calculate_files_size expects
        metadata_file = os.path.join(run_dir, "test_run.json")
        with open(metadata_file, "w") as f:
            json.dump({"metadata": {}}, f)

        record_input(run_dir, run_id="test_run", input_data=input_data)

        # Check that inputs directory was created
        inputs_dir = os.path.join(run_dir, "inputs")
        self.assertTrue(os.path.exists(inputs_dir))

        # Check that input file was created with correct content
        input_file = os.path.join(inputs_dir, "input")
        self.assertTrue(os.path.exists(input_file))

        with open(input_file) as f:
            saved_data = f.read()

        self.assertEqual(saved_data, input_data)

    def test_record_input_with_inputs_dir_path(self):
        """Test record_input with inputs directory path."""
        run_dir = os.path.join(self.test_dir, "test_run")
        os.makedirs(run_dir)

        # Create a test inputs directory with some files
        test_inputs_dir = os.path.join(self.test_dir, "test_inputs")
        os.makedirs(test_inputs_dir)

        with open(os.path.join(test_inputs_dir, "file1.csv"), "w") as f:
            f.write("col1,col2\nval1,val2\n")

        with open(os.path.join(test_inputs_dir, "file2.txt"), "w") as f:
            f.write("test content")

        # Create subdirectory
        subdir = os.path.join(test_inputs_dir, "subdir")
        os.makedirs(subdir)
        with open(os.path.join(subdir, "file3.json"), "w") as f:
            json.dump({"test": "data"}, f)

        # Create minimal metadata file that calculate_files_size expects
        metadata_file = os.path.join(run_dir, "test_run.json")
        with open(metadata_file, "w") as f:
            json.dump({"metadata": {}}, f)

        record_input(run_dir, run_id="test_run", inputs_dir_path=test_inputs_dir)

        # Check that inputs directory was created
        inputs_dir = os.path.join(run_dir, "inputs")
        self.assertTrue(os.path.exists(inputs_dir))

        # Check that all files were copied
        self.assertTrue(os.path.exists(os.path.join(inputs_dir, "file1.csv")))
        self.assertTrue(os.path.exists(os.path.join(inputs_dir, "file2.txt")))
        self.assertTrue(os.path.exists(os.path.join(inputs_dir, "subdir", "file3.json")))

        # Verify content
        with open(os.path.join(inputs_dir, "file1.csv")) as f:
            self.assertEqual(f.read(), "col1,col2\nval1,val2\n")

    def test_record_input_prioritizes_inputs_dir_path(self):
        """Test that inputs_dir_path takes precedence over input_data."""
        run_dir = os.path.join(self.test_dir, "test_run")
        os.makedirs(run_dir)

        # Create test inputs directory
        test_inputs_dir = os.path.join(self.test_dir, "test_inputs")
        os.makedirs(test_inputs_dir)
        with open(os.path.join(test_inputs_dir, "test.txt"), "w") as f:
            f.write("from directory")

        input_data = {"should": "be ignored"}

        # Create minimal metadata file that calculate_files_size expects
        metadata_file = os.path.join(run_dir, "test_run.json")
        with open(metadata_file, "w") as f:
            json.dump({"metadata": {}}, f)

        record_input(run_dir, run_id="test_run", input_data=input_data, inputs_dir_path=test_inputs_dir)

        inputs_dir = os.path.join(run_dir, "inputs")

        # Should have the directory file, not the JSON input
        self.assertTrue(os.path.exists(os.path.join(inputs_dir, "test.txt")))
        self.assertFalse(os.path.exists(os.path.join(inputs_dir, "input.json")))

    def test_record_input_handles_nonexistent_inputs_dir(self):
        """Test record_input handles non-existent inputs directory gracefully."""
        run_dir = os.path.join(self.test_dir, "test_run")
        os.makedirs(run_dir)

        nonexistent_dir = os.path.join(self.test_dir, "nonexistent")

        # Create minimal metadata file that calculate_files_size expects
        metadata_file = os.path.join(run_dir, "test_run.json")
        with open(metadata_file, "w") as f:
            json.dump({"metadata": {}}, f)

        # Should not raise an exception
        record_input(run_dir, run_id="test_run", inputs_dir_path=nonexistent_dir)

        # Inputs directory should still be created
        inputs_dir = os.path.join(run_dir, "inputs")
        self.assertTrue(os.path.exists(inputs_dir))

    @patch("nextmv.local.runner.subprocess.Popen")
    @patch("nextmv.local.runner.safe_id")
    def test_run_function_execution(self, mock_safe_id, mock_popen):
        """Test the main run function execution flow."""
        # Setup mocks
        mock_safe_id.return_value = "test-run-id"
        mock_process = Mock()
        mock_process.stdin = Mock()
        mock_popen.return_value = mock_process

        # Create manifest
        manifest = Manifest(
            files=["main.py"],
            runtime=ManifestRuntime.PYTHON,
        )

        run_config = {"format": {"input": {"type": "json"}, "output": {"type": "json"}}}

        input_data = {"test": "input"}
        options = {"duration": "10s"}

        result = run(
            app_id="sample-app",
            src=self.test_src,
            manifest=manifest,
            run_config=run_config,
            input_data=input_data,
            options=options,
        )

        # Verify run ID was generated
        mock_safe_id.assert_called_once_with("local")
        self.assertEqual(result, "test-run-id")

        # Verify subprocess was called
        mock_popen.assert_called_once()
        popen_args = mock_popen.call_args

        # Check the command
        self.assertEqual(popen_args[0][0], [sys.executable, "executor.py"])

        # Check that stdin was written to
        mock_process.stdin.write.assert_called_once()
        mock_process.stdin.close.assert_called_once()

        # Verify the input JSON that was sent
        stdin_data = mock_process.stdin.write.call_args[0][0]
        stdin_json = json.loads(stdin_data)

        expected_keys = [
            "run_id",
            "src",
            "manifest_dict",
            "run_dir",
            "run_config",
            "input_data",
            "inputs_dir_path",
            "options",
        ]
        for key in expected_keys:
            self.assertIn(key, stdin_json)

        self.assertEqual(stdin_json["input_data"], input_data)
        self.assertEqual(stdin_json["options"], options)
        self.assertEqual(stdin_json["run_config"], run_config)

    @patch("nextmv.local.runner.subprocess.Popen")
    @patch("nextmv.local.runner.safe_id")
    def test_run_with_inputs_dir_path(self, mock_safe_id, mock_popen):
        """Test run function with inputs directory path."""
        mock_safe_id.return_value = "test-run-id-2"
        mock_process = Mock()
        mock_process.stdin = Mock()
        mock_popen.return_value = mock_process

        manifest = Manifest(
            files=["main.py"],
        )

        run_config = {"format": {"input": {"type": "multi-file"}, "output": {"type": "multi-file"}}}

        # Create test inputs directory
        test_inputs_dir = os.path.join(self.test_dir, "test_inputs")
        os.makedirs(test_inputs_dir)

        result = run(
            app_id="sample-app",
            src=self.test_src,
            manifest=manifest,
            run_config=run_config,
            inputs_dir_path=test_inputs_dir,
        )

        self.assertEqual(result, "test-run-id-2")  # Verify the input JSON included the absolute path
        stdin_data = mock_process.stdin.write.call_args[0][0]
        stdin_json = json.loads(stdin_data)

        self.assertEqual(stdin_json["inputs_dir_path"], os.path.abspath(test_inputs_dir))

    @patch("nextmv.local.runner.subprocess.Popen")
    @patch("nextmv.local.runner.safe_id")
    def test_run_with_no_inputs_dir_path(self, mock_safe_id, mock_popen):
        """Test run function when inputs_dir_path is None."""
        mock_safe_id.return_value = "test-run-id-3"
        mock_process = Mock()
        mock_process.stdin = Mock()
        mock_popen.return_value = mock_process

        manifest = Manifest(
            files=["main.py"],
        )

        run_config = {"format": {"input": {"type": "json"}}}

        run(
            app_id="sample-app",
            src=self.test_src,
            manifest=manifest,
            run_config=run_config,
            input_data={"test": "data"},
        )

        # Verify the input JSON has None for inputs_dir_path
        stdin_data = mock_process.stdin.write.call_args[0][0]
        stdin_json = json.loads(stdin_data)

        self.assertIsNone(stdin_json["inputs_dir_path"])

    @patch("nextmv.local.runner.subprocess.Popen")
    @patch("nextmv.local.runner.safe_id")
    def test_run_subprocess_configuration(self, mock_safe_id, mock_popen):
        """Test that subprocess is configured correctly for detached execution."""
        mock_safe_id.return_value = "test-run-id-4"
        mock_process = Mock()
        mock_process.stdin = Mock()
        mock_popen.return_value = mock_process

        manifest = Manifest(
            files=["main.py"],
        )

        run_config = {"format": {"input": {"type": "json"}}}

        run(
            app_id="sample-app",
            src=self.test_src,
            manifest=manifest,
            run_config=run_config,
            input_data={"test": "data"},
        )

        # Verify subprocess configuration
        popen_kwargs = mock_popen.call_args[1]

        self.assertEqual(popen_kwargs["env"], os.environ)
        self.assertTrue(popen_kwargs["text"])
        self.assertEqual(popen_kwargs["stdin"], unittest.mock.ANY)  # subprocess.PIPE
        self.assertEqual(popen_kwargs["stdout"], unittest.mock.ANY)  # subprocess.DEVNULL
        self.assertEqual(popen_kwargs["stderr"], unittest.mock.ANY)  # subprocess.DEVNULL
        self.assertTrue(popen_kwargs["start_new_session"])  # For detached process


if __name__ == "__main__":
    unittest.main()



"""
Module for declarations related to upload and download URLs in Nextmv Cloud.

Classes
-------
DownloadURL
    Represents a download URL for fetching content from Nextmv Cloud.
UploadURL
    Represents an upload URL for sending content to Nextmv Cloud.
"""

from nextmv.base_model import BaseModel


class DownloadURL(BaseModel):
    """
    Result of getting a download URL.

    You can import the `DownloadURL` class directly from `cloud`:

    ```python
    from nextmv.cloud import DownloadURL
    ```

    This class represents a download URL that can be used to fetch content
    from Nextmv Cloud, typically used for downloading large run results.

    Attributes
    ----------
    url : str
        URL to use for downloading the file.

    Examples
    --------
    >>> download_url = DownloadURL(url="https://example.com/download")
    >>> response = requests.get(download_url.url)
    """

    url: str
    """URL to use for downloading the file."""


class UploadURL(BaseModel):
    """
    Result of getting an upload URL.

    You can import the `UploadURL` class directly from `cloud`:

    ```python
    from nextmv.cloud import UploadURL
    ```

    This class represents an upload URL that can be used to send data to
    Nextmv Cloud, typically used for uploading large inputs for runs.

    Attributes
    ----------
    upload_id : str
        ID of the upload, used to reference the uploaded content.
    upload_url : str
        URL to use for uploading the file.

    Examples
    --------
    >>> upload_url = UploadURL(upload_id="123", upload_url="https://example.com/upload")
    >>> with open("large_input.json", "rb") as f:
    ...     requests.put(upload_url.upload_url, data=f)
    """

    upload_id: str
    """ID of the upload."""
    upload_url: str
    """URL to use for uploading the file."""



from enum import Enum

import nextmv


# Define choice enum
class ChoiceEnum(Enum):
    choice1 = "choice1"
    choice2 = "choice2"


options = nextmv.Options(
    nextmv.Option("choice_opt", ChoiceEnum, default=ChoiceEnum.choice1),
)

print(options.to_dict())



"""Classes for working with Nextmv Cloud Instances.

This module provides classes for interacting with instances in Nextmv Cloud.
It defines the core data structures for both instance configuration and the
instance itself.

Classes
-------
InstanceConfiguration
    Configuration settings for a Nextmv Cloud instance.
Instance
    Representation of a Nextmv Cloud instance tied to an application version.
"""

from datetime import datetime

from nextmv.base_model import BaseModel
from nextmv.run import RunQueuing


class InstanceConfiguration(BaseModel):
    """Configuration for a Nextmv Cloud instance.

    You can import the `InstanceConfiguration` class directly from `cloud`:

    ```python
    from nextmv.cloud import InstanceConfiguration
    ```

    This class represents the configuration settings that can be applied to a
    Nextmv Cloud instance, including execution class, options, and secrets.

    Parameters
    ----------
    execution_class : str, optional
        The execution class for the instance, which determines compute resources.
    options : dict, optional
        Runtime options/parameters for the application.
    secrets_collection_id : str, optional
        ID of the secrets collection to use with this instance.
    queuing : RunQueuing, optional
        Queuing configuration for the instance.
    integration_id : str, optional
        ID of the integration to use for the instance.

    Examples
    --------
    >>> config = InstanceConfiguration(
    ...     execution_class="small",
    ...     options={"max_runtime": 30},
    ...     secrets_collection_id="sc_1234567890"
    ... )
    """

    execution_class: str | None = None
    """Execution class for the instance."""
    options: dict | None = None
    """Options of the app that the instance uses."""
    secrets_collection_id: str | None = None
    """ID of the secrets collection that the instance uses."""
    queuing: RunQueuing | None = None
    """Queuing configuration for the instance."""
    integration_id: str | None = None
    """ID of the integration to use for the instance."""

    def model_post_init(self, __context) -> None:
        """
        Validations done after parsing the model.

        Raises
        ------
        ValueError
            If execution_class is an empty string.
        """

        if self.integration_id is None or self.integration_id == "":
            return

        integration_val = "integration"
        if self.execution_class is not None and self.execution_class != "" and self.execution_class != integration_val:
            raise ValueError(f"When integration_id is set, execution_class must be `{integration_val}` or None.")

        self.execution_class = integration_val


class Instance(BaseModel):
    """An instance of an application tied to a version with configuration.

    You can import the `Instance` class directly from `cloud`:

    ```python
    from nextmv.cloud import Instance
    ```

    A Nextmv Cloud instance represents a deployable configuration of an application
    version. Instances have their own unique identity and can be used to run jobs
    with specific configurations.

    Parameters
    ----------
    id : str
        The unique identifier of the instance.
    application_id : str
        ID of the application that this instance belongs to.
    version_id : str
        ID of the application version this instance uses.
    name : str
        Human-readable name of the instance.
    description : str
        Detailed description of the instance.
    configuration : InstanceConfiguration
        Configuration settings for this instance.
    locked : bool
        Whether the instance is locked for modifications.
    created_at : datetime
        Timestamp when the instance was created.
    updated_at : datetime
        Timestamp when the instance was last updated.

    Examples
    --------
    >>> from nextmv.cloud import Instance, InstanceConfiguration
    >>> instance = Instance(
    ...     id="inst_1234567890",
    ...     application_id="app_1234567890",
    ...     version_id="ver_1234567890",
    ...     name="Production Routing Instance",
    ...     description="Instance for daily production routing jobs",
    ...     configuration=InstanceConfiguration(execution_class="small"),
    ...     locked=False,
    ...     created_at=datetime.now(),
    ...     updated_at=datetime.now()
    ... )
    """

    id: str
    """ID of the instance."""
    application_id: str
    """ID of the application that this is an instance of."""
    version_id: str
    """ID of the version that this instance is uses."""
    name: str
    """Name of the instance."""
    description: str
    """Description of the instance."""
    configuration: InstanceConfiguration
    """Configuration for the instance."""
    locked: bool
    """Whether the instance is locked."""
    created_at: datetime
    """Creation time of the instance."""
    updated_at: datetime
    """Last update time of the instance."""



import datetime
import time
import unittest

from nextmv.run import (
    Format,
    FormatInput,
    FormatOutput,
    Metadata,
    Run,
    RunConfiguration,
    RunInformation,
    RunTypeConfiguration,
    run_duration,
)
from nextmv.status import Status, StatusV2


class TestRunDuration(unittest.TestCase):
    def test_run_duration_convenience_func(self):
        before_t, before_dt = time.time(), datetime.datetime.now()
        diff = 0.25
        after_t, after_dt = before_t + diff, before_dt + datetime.timedelta(seconds=diff)
        duration_t = run_duration(before_t, after_t)
        duration_dt = run_duration(before_dt, after_dt)
        self.assertAlmostEqual(duration_t, 250.0, delta=1.0)
        self.assertAlmostEqual(duration_dt, 250.0, delta=1.0)
        self.assertIsInstance(duration_t, int)
        self.assertIsInstance(duration_dt, int)


class TestRunInformationToRun(unittest.TestCase):
    def setUp(self):
        """Set up test data for RunInformation to_run() tests."""
        self.created_at = datetime.datetime(2023, 10, 15, 14, 30, 0)

        # Create a complete metadata object
        self.metadata = Metadata(
            application_id="app-12345",
            application_instance_id="instance-67890",
            application_version_id="version-11111",
            created_at=self.created_at,
            duration=5000.0,
            error="",
            input_size=1024.0,
            output_size=2048.0,
            format=Format(format_input=FormatInput(), format_output=FormatOutput()),
            status_v2=StatusV2.succeeded,
            status=Status.succeeded,
        )

        # Create a RunInformation instance
        self.run_info = RunInformation(
            id="run-98765",
            description="Test optimization run",
            name="Test Run",
            user_email="test@example.com",
            metadata=self.metadata,
            console_url="https://console.nextmv.io/run/run-98765",
            synced_run_id="synced-12345",
            synced_at=datetime.datetime(2023, 10, 15, 15, 0, 0),
        )

    def test_to_run_basic_transformation(self):
        """Test basic transformation from RunInformation to Run."""
        run = self.run_info.to_run()

        # Test that the result is a Run instance
        self.assertIsInstance(run, Run)

        # Test direct mappings from RunInformation
        self.assertEqual(run.id, self.run_info.id)
        self.assertEqual(run.user_email, self.run_info.user_email)
        self.assertEqual(run.name, self.run_info.name)
        self.assertEqual(run.description, self.run_info.description)

        # Test mappings from metadata
        self.assertEqual(run.created_at, self.metadata.created_at)
        self.assertEqual(run.application_id, self.metadata.application_id)
        self.assertEqual(run.application_instance_id, self.metadata.application_instance_id)
        self.assertEqual(run.application_version_id, self.metadata.application_version_id)
        self.assertEqual(run.status, self.metadata.status)
        self.assertEqual(run.status_v2, self.metadata.status_v2)

    def test_to_run_default_values(self):
        """Test that unavailable attributes are set to appropriate defaults."""
        run = self.run_info.to_run()

        # Test default RunTypeConfiguration
        self.assertIsInstance(run.run_type, RunTypeConfiguration)
        self.assertIsNone(run.run_type.run_type)
        self.assertIsNone(run.run_type.definition_id)
        self.assertIsNone(run.run_type.reference_id)

        # Test empty string defaults
        self.assertEqual(run.execution_class, "")
        self.assertEqual(run.runtime, "")

        # Test None defaults for optional fields
        self.assertIsNone(run.queuing_priority)
        self.assertIsNone(run.queuing_disabled)
        self.assertIsNone(run.experiment_id)
        self.assertIsNone(run.statistics)
        self.assertIsNone(run.input_id)
        self.assertIsNone(run.option_set)
        self.assertIsNone(run.options)
        self.assertIsNone(run.request_options)
        self.assertIsNone(run.options_summary)
        self.assertIsNone(run.scenario_id)
        self.assertIsNone(run.repetition)
        self.assertIsNone(run.input_set_id)

    def test_to_run_with_minimal_metadata(self):
        """Test transformation with minimal metadata (only status without deprecated status)."""
        minimal_metadata = Metadata(
            application_id="app-minimal",
            application_instance_id="instance-minimal",
            application_version_id="version-minimal",
            created_at=self.created_at,
            duration=1000.0,
            error="Test error",
            input_size=512.0,
            output_size=1024.0,
            format=Format(format_input=FormatInput(), format_output=FormatOutput()),
            status_v2=StatusV2.failed,
            status=Status.failed,  # Add required status field
        )

        run_info_minimal = RunInformation(
            id="run-minimal",
            description="Minimal test",
            name="Minimal",
            user_email="minimal@example.com",
            metadata=minimal_metadata,
        )

        run = run_info_minimal.to_run()

        # Test that transformation works with minimal data
        self.assertEqual(run.id, "run-minimal")
        self.assertEqual(run.application_id, "app-minimal")
        self.assertEqual(run.status_v2, StatusV2.failed)
        self.assertEqual(run.status, Status.failed)  # Status is now provided

    def test_to_run_preserves_datetime_objects(self):
        """Test that datetime objects are properly preserved during transformation."""
        run = self.run_info.to_run()

        # Test that datetime objects are preserved and are the same instance
        self.assertIs(run.created_at, self.metadata.created_at)
        self.assertEqual(run.created_at, self.created_at)
        self.assertIsInstance(run.created_at, datetime.datetime)

    def test_to_run_multiple_transformations(self):
        """Test that multiple transformations from the same RunInformation work correctly."""
        run1 = self.run_info.to_run()
        run2 = self.run_info.to_run()

        # Test that both transformations produce equivalent results
        self.assertEqual(run1.id, run2.id)
        self.assertEqual(run1.user_email, run2.user_email)
        self.assertEqual(run1.name, run2.name)
        self.assertEqual(run1.description, run2.description)
        self.assertEqual(run1.created_at, run2.created_at)
        self.assertEqual(run1.application_id, run2.application_id)
        self.assertEqual(run1.status_v2, run2.status_v2)

        # Test that they are different instances
        self.assertIsNot(run1, run2)
        self.assertIsNot(run1.run_type, run2.run_type)

    def test_to_run_with_different_status_values(self):
        """Test transformation with different status values."""
        # Test with different StatusV2 values
        status_values = [StatusV2.succeeded, StatusV2.failed, StatusV2.running, StatusV2.queued]

        for status_v2 in status_values:
            with self.subTest(status_v2=status_v2):
                # Map StatusV2 to Status for compatibility
                status_mapping = {
                    StatusV2.succeeded: Status.succeeded,
                    StatusV2.failed: Status.failed,
                    StatusV2.running: Status.running,
                    StatusV2.queued: Status.running,  # Map queued to running since Status doesn't have queued
                }
                metadata = Metadata(
                    application_id="app-test",
                    application_instance_id="instance-test",
                    application_version_id="version-test",
                    created_at=self.created_at,
                    duration=2000.0,
                    error="",
                    input_size=256.0,
                    output_size=512.0,
                    format=Format(format_input=FormatInput(), format_output=FormatOutput()),
                    status_v2=status_v2,
                    status=status_mapping[status_v2],
                )

                run_info = RunInformation(
                    id=f"run-{status_v2.value}",
                    description=f"Test with {status_v2.value}",
                    name="Status Test",
                    user_email="status@example.com",
                    metadata=metadata,
                )

                run = run_info.to_run()
                self.assertEqual(run.status_v2, status_v2)


class TestRunConfigurationValidation(unittest.TestCase):
    """Test validation logic in RunConfiguration.model_post_init."""

    def test_no_integration_id_no_validation(self):
        """Test that validation is skipped when integration_id is None or empty."""
        # With None integration_id
        config = RunConfiguration(integration_id=None, execution_class="small")
        self.assertEqual(config.execution_class, "small")

        # With empty string integration_id
        config = RunConfiguration(integration_id="", execution_class="large")
        self.assertEqual(config.execution_class, "large")

        # With no integration_id specified
        config = RunConfiguration(execution_class="medium")
        self.assertEqual(config.execution_class, "medium")

    def test_integration_id_with_integration_execution_class(self):
        """Test that integration_id with execution_class='integration' is valid."""
        config = RunConfiguration(integration_id="int-12345", execution_class="integration")
        self.assertEqual(config.integration_id, "int-12345")
        self.assertEqual(config.execution_class, "integration")

    def test_integration_id_with_none_execution_class(self):
        """Test that integration_id with execution_class=None sets it to 'integration'."""
        config = RunConfiguration(integration_id="int-12345", execution_class=None)
        self.assertEqual(config.integration_id, "int-12345")
        self.assertEqual(config.execution_class, "integration")

    def test_integration_id_without_execution_class(self):
        """Test that integration_id without execution_class sets it to 'integration'."""
        config = RunConfiguration(integration_id="int-12345")
        self.assertEqual(config.integration_id, "int-12345")
        self.assertEqual(config.execution_class, "integration")

    def test_integration_id_with_empty_execution_class(self):
        """Test that integration_id with execution_class='' sets it to 'integration'."""
        config = RunConfiguration(integration_id="int-12345", execution_class="")
        self.assertEqual(config.integration_id, "int-12345")
        self.assertEqual(config.execution_class, "integration")

    def test_integration_id_with_invalid_execution_class_raises_error(self):
        """Test that integration_id with non-integration execution_class raises ValueError."""
        invalid_classes = ["small", "medium", "large", "custom", "standard"]

        for execution_class in invalid_classes:
            with self.subTest(execution_class=execution_class):
                with self.assertRaises(ValueError) as context:
                    RunConfiguration(integration_id="int-12345", execution_class=execution_class)

                error_msg = str(context.exception)
                self.assertIn("When integration_id is set", error_msg)
                self.assertIn("execution_class must be `integration` or None", error_msg)

    def test_integration_id_error_message_format(self):
        """Test that the error message contains the expected format."""
        with self.assertRaises(ValueError) as context:
            RunConfiguration(integration_id="int-12345", execution_class="custom")

        error_msg = str(context.exception)
        # When using model_post_init, Pydantic wraps the error message
        self.assertIn("When integration_id is set, execution_class must be `integration` or None.", error_msg)



"""
Unit tests for the nextmv configuration command.
"""

import os
import unittest
from pathlib import Path
from unittest.mock import mock_open, patch

from nextmv.cli.configuration.config import (
    API_KEY_KEY,
    CONFIG_DIR,
    CONFIG_FILE,
    DEFAULT_ENDPOINT,
    ENDPOINT_KEY,
    GO_CLI_PATH,
    load_config,
    obscure_api_key,
    save_config,
)
from nextmv.cli.main import app, check_config_in_path, go_cli_exists, remove_go_cli
from typer.testing import CliRunner


class TestConfigureCommand(unittest.TestCase):
    """Tests for the configure CLI command."""

    def setUp(self):
        self.runner = CliRunner()
        self.app = app
        self.test_api_key = "test_api_key_12345"
        self.test_profile = "test_profile"

        # Backup and clear NEXTMV_API_KEY env var if set, this is needed for
        # tests that check its absence.
        env_api_key = os.getenv("NEXTMV_API_KEY")
        self.env_api_key = env_api_key
        if env_api_key is not None:
            del os.environ["NEXTMV_API_KEY"]

    def tearDown(self):
        # Restore NEXTMV_API_KEY env var if it was set.
        if self.env_api_key is not None:
            os.environ["NEXTMV_API_KEY"] = self.env_api_key

    def test_configure_no_args_shows_error(self):
        """Test that running configuration create without arguments shows an error."""
        result = self.runner.invoke(self.app, ["configuration", "create"])
        self.assertEqual(result.exit_code, 2)

    @patch("nextmv.cli.configuration.create.save_config")
    @patch("nextmv.cli.configuration.create.load_config")
    def test_configure_default_profile(self, mock_load, mock_save):
        """Test configuring the default profile with an API key."""
        mock_load.return_value = {}

        result = self.runner.invoke(self.app, ["configuration", "create", "--api-key", self.test_api_key])

        self.assertEqual(result.exit_code, 0)
        self.assertIn("Configuration saved successfully", result.output)
        self.assertIn("Default", result.output)

        # Verify save was called with the correct config
        mock_save.assert_called_once()
        saved_config = mock_save.call_args[0][0]
        self.assertEqual(saved_config[API_KEY_KEY], self.test_api_key)
        self.assertEqual(saved_config[ENDPOINT_KEY], DEFAULT_ENDPOINT)

    @patch("nextmv.cli.configuration.create.save_config")
    @patch("nextmv.cli.configuration.create.load_config")
    def test_configure_named_profile(self, mock_load, mock_save):
        """Test configuring a named profile."""
        mock_load.return_value = {}

        result = self.runner.invoke(
            self.app, ["configuration", "create", "--api-key", self.test_api_key, "--profile", self.test_profile]
        )

        self.assertEqual(result.exit_code, 0)
        self.assertIn("Configuration saved successfully", result.output)
        self.assertIn(self.test_profile, result.output)

        # Verify save was called with the correct config
        mock_save.assert_called_once()
        saved_config = mock_save.call_args[0][0]
        self.assertIn(self.test_profile, saved_config)
        self.assertEqual(saved_config[self.test_profile][API_KEY_KEY], self.test_api_key)

    @patch("nextmv.cli.configuration.create.save_config")
    @patch("nextmv.cli.configuration.create.load_config")
    def test_configure_reserved_profile_name_default(self, mock_load, mock_save):
        """Test that 'default' profile name is rejected."""
        mock_load.return_value = {}

        result = self.runner.invoke(
            self.app, ["configuration", "create", "--api-key", self.test_api_key, "--profile", "default"]
        )

        self.assertEqual(result.exit_code, 1)
        self.assertIn("reserved", result.output.lower())
        mock_save.assert_not_called()

    @patch("nextmv.cli.configuration.list.load_config")
    def test_profiles_empty(self, mock_load):
        """Test showing profiles when no configuration exists."""
        mock_load.return_value = {}

        result = self.runner.invoke(self.app, ["configuration", "list"])

        self.assertEqual(result.exit_code, 1)

    @patch("nextmv.cli.configuration.list.load_config")
    def test_show_profiles_with_config(self, mock_load):
        """Test showing profiles when configuration exists."""
        mock_load.return_value = {
            API_KEY_KEY: "default_key_123",
            ENDPOINT_KEY: DEFAULT_ENDPOINT,
            "custom_profile": {
                API_KEY_KEY: "custom_key_456",
                ENDPOINT_KEY: "custom.endpoint.com",
            },
        }

        result = self.runner.invoke(self.app, ["configuration", "list"])

        self.assertEqual(result.exit_code, 0)
        self.assertIn("Default", result.output)
        self.assertIn("custom_profile", result.output)

    @patch("nextmv.cli.configuration.delete.save_config")
    @patch("nextmv.cli.configuration.delete.load_config")
    @patch("rich.prompt.Confirm.ask", return_value=True)
    def test_delete_profile(self, mock_confirm, mock_load, mock_save):
        """Test deleting a profile."""
        mock_load.return_value = {
            API_KEY_KEY: "default_key",
            ENDPOINT_KEY: DEFAULT_ENDPOINT,
            self.test_profile: {
                API_KEY_KEY: "profile_key",
                ENDPOINT_KEY: DEFAULT_ENDPOINT,
            },
        }

        result = self.runner.invoke(self.app, ["configuration", "delete", "--profile", self.test_profile])

        self.assertEqual(result.exit_code, 0)
        self.assertIn("deleted successfully", result.output)

        # Verify save was called with the profile removed
        mock_save.assert_called_once()
        saved_config = mock_save.call_args[0][0]
        self.assertNotIn(self.test_profile, saved_config)

    @patch("nextmv.cli.configuration.delete.load_config")
    def test_delete_nonexistent_profile(self, mock_load):
        """Test deleting a profile that doesn't exist."""
        mock_load.return_value = {
            API_KEY_KEY: "default_key",
            ENDPOINT_KEY: DEFAULT_ENDPOINT,
        }

        result = self.runner.invoke(self.app, ["configuration", "delete", "--profile", "nonexistent"])

        self.assertEqual(result.exit_code, 1)
        self.assertIn("does not exist", result.output)

    def test_delete_without_profile_shows_error(self):
        """Test that delete subcommand without --profile shows an error."""
        result = self.runner.invoke(self.app, ["configuration", "delete"])

        self.assertEqual(result.exit_code, 2)

    @patch("nextmv.cli.configuration.create.save_config")
    @patch("nextmv.cli.configuration.create.load_config")
    def test_configure_strips_https_from_endpoint(self, mock_load, mock_save):
        """Test that https:// is stripped from the endpoint."""
        mock_load.return_value = {}

        result = self.runner.invoke(
            self.app,
            ["configuration", "create", "--api-key", self.test_api_key, "--endpoint", "https://custom.api.com"],
        )

        self.assertEqual(result.exit_code, 0)

        # Verify the endpoint was saved without https://
        mock_save.assert_called_once()
        saved_config = mock_save.call_args[0][0]
        self.assertEqual(saved_config[ENDPOINT_KEY], "custom.api.com")


class TestObscureApiKey(unittest.TestCase):
    """Tests for the obscure_api_key helper function."""

    def test_obscure_short_key(self):
        """Test obscuring a short API key (4 chars or less)."""
        self.assertEqual(obscure_api_key("ab"), "**")
        self.assertEqual(obscure_api_key("abc"), "***")
        self.assertEqual(obscure_api_key("abcd"), "****")

    def test_obscure_normal_key(self):
        """Test obscuring a normal length API key."""
        result = obscure_api_key("abcdefghij")
        self.assertEqual(result, "ab****ij")
        self.assertEqual(len(result), 8)


class TestLoadConfig(unittest.TestCase):
    """Tests for the load_config helper function."""

    @patch.object(Path, "exists")
    def test_load_config_no_file(self, mock_exists):
        """Test loading config when file doesn't exist."""
        mock_exists.return_value = False

        result = load_config()
        self.assertEqual(result, {})

    @patch.object(Path, "open", mock_open(read_data="apikey: test_key\nendpoint: api.test.com\n"))
    @patch.object(Path, "exists")
    def test_load_config_with_file(self, mock_exists):
        """Test loading config from existing file."""
        mock_exists.return_value = True

        result = load_config()
        self.assertEqual(result[API_KEY_KEY], "test_key")
        self.assertEqual(result[ENDPOINT_KEY], "api.test.com")


class TestSaveConfig(unittest.TestCase):
    """Tests for the save_config helper function."""

    @patch.object(Path, "open", mock_open())
    @patch.object(Path, "mkdir")
    def test_save_config_creates_directory(self, mock_mkdir):
        """Test that save_config creates the config directory."""
        config = {API_KEY_KEY: "test_key", ENDPOINT_KEY: "api.test.com"}
        save_config(config)

        mock_mkdir.assert_called_once_with(parents=True, exist_ok=True)


class TestConfigConstants(unittest.TestCase):
    """Tests for configuration constants."""

    def test_config_dir_is_in_home(self):
        """Test that CONFIG_DIR is in the home directory."""
        self.assertEqual(CONFIG_DIR.parent, Path.home())
        self.assertEqual(CONFIG_DIR.name, ".nextmv")

    def test_config_file_is_yaml(self):
        """Test that CONFIG_FILE is a yaml file."""
        self.assertEqual(CONFIG_FILE.suffix, ".yaml")
        self.assertEqual(CONFIG_FILE.name, "config.yaml")

    def test_default_endpoint(self):
        """Test the default endpoint value."""
        self.assertEqual(DEFAULT_ENDPOINT, "api.cloud.nextmv.io")


class TestGoCliExists(unittest.TestCase):
    """Tests for the go_cli_exists function."""

    @patch("nextmv.cli.main.rich.print")
    @patch("nextmv.cli.main.check_config_in_path")
    @patch.object(Path, "exists")
    def test_go_cli_exists_returns_true_when_file_exists(self, mock_exists, mock_check_path, mock_print):
        """Test that go_cli_exists returns True when the Go CLI file exists."""
        mock_exists.return_value = True

        result = go_cli_exists()

        self.assertTrue(result)
        mock_check_path.assert_called_once()

    @patch("nextmv.cli.main.check_config_in_path")
    @patch.object(Path, "exists")
    def test_go_cli_exists_returns_false_when_file_not_exists(self, mock_exists, mock_check_path):
        """Test that go_cli_exists returns False when the Go CLI file does not exist."""
        mock_exists.return_value = False

        result = go_cli_exists()

        self.assertFalse(result)
        mock_check_path.assert_called_once()


class TestRemoveGoCli(unittest.TestCase):
    """Tests for the remove_go_cli function."""

    @patch("nextmv.cli.main.rich.print")
    @patch("nextmv.cli.main.check_config_in_path")
    @patch.object(Path, "unlink")
    @patch.object(Path, "exists")
    def test_remove_go_cli_deletes_file_when_exists(self, mock_exists, mock_unlink, mock_check_path, mock_print):
        """Test that remove_go_cli deletes the file when it exists."""
        mock_exists.return_value = True

        remove_go_cli()

        mock_unlink.assert_called_once()
        mock_check_path.assert_called_once()

    @patch("nextmv.cli.main.check_config_in_path")
    @patch.object(Path, "unlink")
    @patch.object(Path, "exists")
    def test_remove_go_cli_does_not_delete_when_not_exists(self, mock_exists, mock_unlink, mock_check_path):
        """Test that remove_go_cli does not attempt to delete when file does not exist."""
        mock_exists.return_value = False

        remove_go_cli()

        mock_unlink.assert_not_called()
        mock_check_path.assert_called_once()


class TestCheckConfigInPath(unittest.TestCase):
    """Tests for the check_config_in_path function."""

    @patch("nextmv.cli.main.rich.print")
    @patch.dict(os.environ, {"PATH": f"/usr/bin{os.pathsep}{CONFIG_DIR}{os.pathsep}/usr/local/bin"})
    def test_check_config_in_path_prints_warning_when_in_path(self, mock_print):
        """Test that a warning is printed when CONFIG_DIR is in PATH."""
        check_config_in_path()

        mock_print.assert_called_once()
        call_args = str(mock_print.call_args)
        self.assertIn("PATH", call_args)

    @patch("nextmv.cli.main.rich.print")
    @patch.dict(os.environ, {"PATH": "/usr/bin:/usr/local/bin"})
    def test_check_config_in_path_no_warning_when_not_in_path(self, mock_print):
        """Test that no warning is printed when CONFIG_DIR is not in PATH."""
        check_config_in_path()

        mock_print.assert_not_called()


class TestGoCliPath(unittest.TestCase):
    """Tests for the GO_CLI_PATH constant."""

    def test_go_cli_path_is_in_config_dir(self):
        """Test that GO_CLI_PATH is inside CONFIG_DIR."""
        self.assertEqual(GO_CLI_PATH.parent, CONFIG_DIR)
        self.assertEqual(GO_CLI_PATH.name, "nextmv")


if __name__ == "__main__":
    unittest.main()



"""Nextmv Python SDK."""

from .__about__ import __version__
from .base_model import BaseModel as BaseModel
from .base_model import from_dict as from_dict
from .input import DataFile as DataFile
from .input import Input as Input
from .input import InputFormat as InputFormat
from .input import InputLoader as InputLoader
from .input import LocalInputLoader as LocalInputLoader
from .input import csv_data_file as csv_data_file
from .input import json_data_file as json_data_file
from .input import load as load
from .input import load_local as load_local
from .input import text_data_file as text_data_file
from .logger import log as log
from .logger import redirect_stdout as redirect_stdout
from .logger import reset_stdout as reset_stdout
from .manifest import MANIFEST_FILE_NAME as MANIFEST_FILE_NAME
from .manifest import Manifest as Manifest
from .manifest import ManifestBuild as ManifestBuild
from .manifest import ManifestExecution as ManifestExecution
from .manifest import ManifestOption as ManifestOption
from .manifest import ManifestPython as ManifestPython
from .manifest import ManifestPythonArch as ManifestPythonArch
from .manifest import ManifestPythonModel as ManifestPythonModel
from .manifest import ManifestRuntime as ManifestRuntime
from .manifest import ManifestType as ManifestType
from .manifest import default_python_manifest as default_python_manifest
from .model import Model as Model
from .model import ModelConfiguration as ModelConfiguration
from .options import Option as Option
from .options import Options as Options
from .options import Parameter as Parameter
from .output import Asset as Asset
from .output import DataPoint as DataPoint
from .output import LocalOutputWriter as LocalOutputWriter
from .output import Output as Output
from .output import OutputFormat as OutputFormat
from .output import OutputWriter as OutputWriter
from .output import ResultStatistics as ResultStatistics
from .output import RunStatistics as RunStatistics
from .output import Series as Series
from .output import SeriesData as SeriesData
from .output import SolutionFile as SolutionFile
from .output import Statistics as Statistics
from .output import Visual as Visual
from .output import VisualSchema as VisualSchema
from .output import csv_solution_file as csv_solution_file
from .output import json_solution_file as json_solution_file
from .output import text_solution_file as text_solution_file
from .output import write as write
from .output import write_local as write_local
from .polling import DEFAULT_POLLING_OPTIONS as DEFAULT_POLLING_OPTIONS
from .polling import PollingOptions as PollingOptions
from .polling import poll as poll
from .run import ErrorLog as ErrorLog
from .run import ExternalRunResult as ExternalRunResult
from .run import Format as Format
from .run import FormatInput as FormatInput
from .run import FormatOutput as FormatOutput
from .run import Metadata as Metadata
from .run import OptionsSummaryItem as OptionsSummaryItem
from .run import Run as Run
from .run import RunConfiguration as RunConfiguration
from .run import RunInformation as RunInformation
from .run import RunInfoStatistics as RunInfoStatistics
from .run import RunLog as RunLog
from .run import RunQueuing as RunQueuing
from .run import RunResult as RunResult
from .run import RunType as RunType
from .run import RunTypeConfiguration as RunTypeConfiguration
from .run import StatisticsIndicator as StatisticsIndicator
from .run import SyncedRun as SyncedRun
from .run import TrackedRun as TrackedRun
from .run import TrackedRunStatus as TrackedRunStatus
from .run import run_duration as run_duration
from .safe import safe_id as safe_id
from .safe import safe_name_and_id as safe_name_and_id
from .status import Status as Status
from .status import StatusV2 as StatusV2

VERSION = __version__
"""The version of the Nextmv Python SDK."""



"""
This module defines the configuration command tree for the Nextmv CLI.
"""

import typer

from nextmv.cli.configuration.create import app as create_app
from nextmv.cli.configuration.delete import app as delete_app
from nextmv.cli.configuration.list import app as list_app

# Set up subcommand application.
app = typer.Typer()
app.add_typer(create_app)
app.add_typer(delete_app)
app.add_typer(list_app)


@app.callback()
def callback() -> None:
    """
    Configure the CLI and manage profiles.
    """
    pass



"""Manages application versions within the Nextmv Cloud API.

This module provides data models for representing application versions,
their executables, and associated requirements. These models are used
for interacting with version-related endpoints of the Nextmv Cloud API,
allowing users to define, retrieve, and manage different versions of their
decision applications.

Classes
-------
VersionExecutableRequirements
    Defines the requirements for a version's executable, such as type and runtime.
VersionExecutable
    Represents the executable artifact for a specific application version.
Version
    Represents a version of an application, linking to its executable and metadata.
"""

from datetime import datetime

from nextmv.base_model import BaseModel


class VersionExecutableRequirements(BaseModel):
    """
    Requirements for a version executable.

    You can import the `VersionExecutableRequirements` class directly from `cloud`:

    ```python
    from nextmv.cloud import VersionExecutableRequirements
    ```

    These requirements specify the environment and type of executable needed
    to run a particular version of an application.

    Parameters
    ----------
    executable_type : str
        The type of the executable (e.g., "binary", "docker").
    runtime : str
        The runtime environment for the executable (e.g., "go", "python").

    Examples
    --------
    >>> requirements = VersionExecutableRequirements(
    ...     executable_type="binary",
    ...     runtime="go1.x"
    ... )
    >>> print(requirements.executable_type)
    binary
    """

    executable_type: str
    """Type of the executable."""
    runtime: str
    """Runtime for the executable."""


class VersionExecutable(BaseModel):
    """
    Executable for a version.

    You can import the `VersionExecutable` class directly from `cloud`:

    ```python
    from nextmv.cloud import VersionExecutable
    ```

    This class holds information about the actual executable file or image
    associated with an application version, including who uploaded it and when.

    Parameters
    ----------
    id : str
        Unique identifier for the version executable.
    user_email : str
        Email of the user who uploaded the executable.
    uploaded_at : datetime
        Timestamp indicating when the executable was uploaded.
    requirements : VersionExecutableRequirements
        The specific requirements for this executable.

    Examples
    --------
    >>> from datetime import datetime
    >>> reqs = VersionExecutableRequirements(executable_type="docker", runtime="custom")
    >>> executable = VersionExecutable(
    ...     id="exec-123",
    ...     user_email="user@example.com",
    ...     uploaded_at=datetime.now(),
    ...     requirements=reqs
    ... )
    >>> print(executable.id)
    exec-123
    """

    id: str
    """ID of the version."""
    user_email: str
    """Email of the user who uploaded the executable."""
    uploaded_at: datetime
    """Time the executable was uploaded."""
    requirements: VersionExecutableRequirements
    """Requirements for the executable."""


class Version(BaseModel):
    """
    A version of an application representing a code artifact or a compiled binary.

    You can import the `Version` class directly from `cloud`:

    ```python
    from nextmv.cloud import Version
    ```

    This class encapsulates all details of a specific version of an application,
    including its metadata, associated executable, and timestamps.

    Parameters
    ----------
    id : str
        Unique identifier for the version.
    application_id : str
        Identifier of the application to which this version belongs.
    name : str
        User-defined name for the version (e.g., "v1.0.0", "feature-branch-build").
    description : str
        A more detailed description of the version and its changes.
    executable : VersionExecutable
        The executable artifact associated with this version.
    created_at : datetime
        Timestamp indicating when the version was created.
    updated_at : datetime
        Timestamp indicating when the version was last updated.

    Examples
    --------
    >>> from datetime import datetime
    >>> reqs = VersionExecutableRequirements(executable_type="binary", runtime="java11")
    >>> exe = VersionExecutable(
    ...     id="exec-abc",
    ...     user_email="dev@example.com",
    ...     uploaded_at=datetime.now(),
    ...     requirements=reqs
    ... )
    >>> version_info = Version(
    ...     id="ver-xyz",
    ...     application_id="app-123",
    ...     name="Initial Release",
    ...     description="First stable release of the model.",
    ...     executable=exe,
    ...     created_at=datetime.now(),
    ...     updated_at=datetime.now()
    ... )
    >>> print(version_info.name)
    Initial Release
    """

    id: str
    """ID of the version."""
    application_id: str
    """ID of the application that this is a version of."""
    name: str
    """Name of the version."""
    description: str
    """Description of the version."""
    executable: VersionExecutable
    """Executable for the version."""
    created_at: datetime
    """Creation time of the version."""
    updated_at: datetime
    """Last update time of the version."""



import unittest

import nextmv


class TestLogger(unittest.TestCase):
    def test_version(self):
        exported_version = nextmv.VERSION
        expected_version = nextmv.__about__.__version__
        self.assertEqual(exported_version, expected_version)
